import sys
import toml
import os
import shutil
import allencell_ml_segmenter.scripts.bumpver_handler as bumpver_handler
from pathlib import Path

"""
The tests in this file are a bit unconventional. Continue reading for an explanation.

The purpose of bumpver_handler.py is to handle the logic around the args we need to pass
bumpver--the args vary based on which version component we want to bump AND the current version.
This logic is too complex for GH workflows to handle, so we created bumpver_handler.py, which
is called by GH workflows.

We want to verify that the args we pass to bumpver actually make the version updates we expect.
This is tricky to test because bumpver is a tool that reads version state from the repo files and
writes updates to files in the repo, but we don't actually want to change our repo's version over the
course of testing.

The solution here is that we create a 'fake' repo (meaning a fake pyproject.toml) in the _tests/scripts
directory. Then when it comes time to run the bumpver_handler tests, we change our working directory
to _tests/scripts, which tricks bumpver into thinking that our 'fake' pyproject.toml is the source 
of truth.

So, the tests work as follows:
  - change working directory to _tests/scripts
  - copy unmodified_pyproject.toml (a frozen pyproject file that won't be touched by bumpver) to
    pyproject.toml, which resets the version to 0.0.1
  - run bumpver_handler.py, changing sys.argv to modify the args we provide it
  - read pyproject.toml to make sure the version updated as we expected
  - change working directory back to the root of the repo

The first and last steps are accomplished with navigate_to_test_dir_and_reset_version
and navigate_back_to_root_dir, respectively
"""
DEFAULT_START_VERSION: str = "0.0.1"


# tests in this file must call navigate_to_test_dir_and_reset_version at the beginning and navigate_back_to_root_dir at the end
def navigate_to_test_dir_and_reset_version() -> None:
    os.chdir(Path(__file__).parent)
    shutil.copyfile("unmodified_pyproject.toml", "pyproject.toml")


def navigate_back_to_root_dir() -> None:
    os.chdir("../../../..")


def assert_curr_version_number(version: str) -> None:
    curr_version: str = toml.load("pyproject.toml")["project"]["version"]
    assert curr_version == version


def test_bump_patch() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT
    sys.argv = ["bumpver_handler.py", "patch"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("0.0.2")
    # ACT
    sys.argv = ["bumpver_handler.py", "patch"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("0.0.3")
    navigate_back_to_root_dir()


def test_bump_minor() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT
    sys.argv = ["bumpver_handler.py", "minor"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("0.1.0")
    # ACT
    sys.argv = ["bumpver_handler.py", "minor"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("0.2.0")
    navigate_back_to_root_dir()


def test_bump_major() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT
    sys.argv = ["bumpver_handler.py", "major"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("1.0.0")
    # ACT
    sys.argv = ["bumpver_handler.py", "major"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("2.0.0")
    navigate_back_to_root_dir()


def test_bump_rc() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT (create new patch version with .rc0 tag)
    sys.argv = ["bumpver_handler.py", "rc"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("0.0.2rc0")
    # ACT
    sys.argv = ["bumpver_handler.py", "rc"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("0.0.2rc1")
    # ACT (finalize the new patch version)
    sys.argv = ["bumpver_handler.py", "patch"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("0.0.2")
    navigate_back_to_root_dir()


def test_bump_post() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT (create new post version with .post0 tag)
    sys.argv = ["bumpver_handler.py", "post"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("0.0.1post0")
    # ACT
    sys.argv = ["bumpver_handler.py", "post"]
    bumpver_handler.main()
    # ASSERT
    assert_curr_version_number("0.0.1post1")
    navigate_back_to_root_dir()


def test_bump_minor_fails_when_rc_current() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT (create new rc version with .rc0 tag)
    sys.argv = ["bumpver_handler.py", "rc"]
    bumpver_handler.main()
    # ASSERT (sanity check)
    assert_curr_version_number("0.0.2rc0")

    got_expected_exception: bool = False
    try:
        sys.argv = ["bumpver_handler.py", "minor"]
        bumpver_handler.main()
    except ValueError as e:
        got_expected_exception = True
    except:
        pass

    # ASSERT
    assert got_expected_exception
    navigate_back_to_root_dir()


def test_bump_major_fails_when_rc_current() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT (create new rc version with .rc0 tag)
    sys.argv = ["bumpver_handler.py", "rc"]
    bumpver_handler.main()
    # ASSERT (sanity check)
    assert_curr_version_number("0.0.2rc0")

    got_expected_exception: bool = False
    try:
        sys.argv = ["bumpver_handler.py", "major"]
        bumpver_handler.main()
    except ValueError as e:
        got_expected_exception = True
    except:
        pass

    # ASSERT
    assert got_expected_exception
    navigate_back_to_root_dir()


def test_bump_post_fails_when_rc_current() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT (create new rc version with .rc0 tag)
    sys.argv = ["bumpver_handler.py", "rc"]
    bumpver_handler.main()
    # ASSERT (sanity check)
    assert_curr_version_number("0.0.2rc0")

    got_expected_exception: bool = False
    try:
        sys.argv = ["bumpver_handler.py", "post"]
        bumpver_handler.main()
    except ValueError as e:
        got_expected_exception = True
    except:
        pass

    # ASSERT
    assert got_expected_exception
    navigate_back_to_root_dir()


def test_bump_minor_fails_when_post_current() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT (create new post version with .post0 tag)
    sys.argv = ["bumpver_handler.py", "post"]
    bumpver_handler.main()
    # ASSERT (sanity check)
    assert_curr_version_number("0.0.1post0")

    got_expected_exception: bool = False
    try:
        sys.argv = ["bumpver_handler.py", "minor"]
        bumpver_handler.main()
    except ValueError as e:
        got_expected_exception = True
    except:
        pass

    # ASSERT
    assert got_expected_exception
    navigate_back_to_root_dir()


def test_bump_major_fails_when_post_current() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT (create new post version with .post0 tag)
    sys.argv = ["bumpver_handler.py", "post"]
    bumpver_handler.main()
    # ASSERT (sanity check)
    assert_curr_version_number("0.0.1post0")

    got_expected_exception: bool = False
    try:
        sys.argv = ["bumpver_handler.py", "major"]
        bumpver_handler.main()
    except ValueError as e:
        got_expected_exception = True
    except:
        pass

    # ASSERT
    assert got_expected_exception
    navigate_back_to_root_dir()


def test_bump_patch_fails_when_post_current() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT (create new post version with .post0 tag)
    sys.argv = ["bumpver_handler.py", "post"]
    bumpver_handler.main()
    # ASSERT (sanity check)
    assert_curr_version_number("0.0.1post0")

    got_expected_exception: bool = False
    try:
        sys.argv = ["bumpver_handler.py", "patch"]
        bumpver_handler.main()
    except ValueError as e:
        got_expected_exception = True
    except:
        pass

    # ASSERT
    assert got_expected_exception
    navigate_back_to_root_dir()


def test_bump_rc_fails_when_post_current() -> None:
    navigate_to_test_dir_and_reset_version()
    # ASSERT (sanity check)
    assert_curr_version_number(DEFAULT_START_VERSION)
    # ACT (create new post version with .post0 tag)
    sys.argv = ["bumpver_handler.py", "post"]
    bumpver_handler.main()
    # ASSERT (sanity check)
    assert_curr_version_number("0.0.1post0")

    got_expected_exception: bool = False
    try:
        sys.argv = ["bumpver_handler.py", "rc"]
        bumpver_handler.main()
    except ValueError as e:
        got_expected_exception = True
    except:
        pass

    # ASSERT
    assert got_expected_exception
    # For version control niceness, we should have this at the end of the
    # last test in this file
    navigate_to_test_dir_and_reset_version()
    navigate_back_to_root_dir()



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Any

from allencell_ml_segmenter.core.image_data_extractor import ImageData


class IImageDataExtractor(ABC):
    def __init__(self) -> None:
        raise RuntimeError(
            "Cannot initialize new singleton, please use .global_instance() instead"
        )

    @abstractmethod
    def extract_image_data(
        self,
        img_path: Path,
        channel: int = 0,
        dims: bool = True,
        np_data: bool = True,
        seg: Optional[int] = None,
    ) -> ImageData:
        pass

    @classmethod
    @abstractmethod
    def global_instance(cls) -> Any:
        pass



from abc import abstractmethod
from pathlib import Path
from typing import Optional

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.publisher import Publisher


class IExperimentsModel(Publisher):
    """
    Interface for implementing and testing ExperimentsModel
    """

    def __init__(self) -> None:
        super().__init__()

        # state
        self._experiment_name: Optional[str] = None

    def get_experiment_name(self) -> Optional[str]:
        """
        Gets experiment name
        """
        return self._experiment_name

    def apply_experiment_name(self, name: Optional[str]) -> None:
        """
        Sets experiment name

        name (str): name of cyto-dl experiment
        """
        print(f"applying {name} experiment")
        self._experiment_name = name
        self.dispatch(Event.ACTION_EXPERIMENT_APPLIED)

    @abstractmethod
    def get_best_ckpt(self) -> Optional[Path]:
        pass

    @abstractmethod
    def get_experiments(self) -> list[str]:
        pass

    @abstractmethod
    def refresh_experiments(self) -> None:
        pass

    @abstractmethod
    def get_user_experiments_path(self) -> Optional[Path]:
        pass

    @abstractmethod
    def get_model_checkpoints_path(
        self, experiment_name: Optional[str], checkpoint: Optional[str]
    ) -> Path:
        pass

    @abstractmethod
    def get_metrics_csv_path(self) -> Optional[Path]:
        pass

    @abstractmethod
    def get_latest_metrics_csv_version(self) -> int:
        pass

    @abstractmethod
    def get_latest_metrics_csv_path(self) -> Optional[Path]:
        pass

    @abstractmethod
    def get_csv_path(self) -> Optional[Path]:
        pass

    @abstractmethod
    def get_cache_dir(self) -> Optional[Path]:
        pass

    @abstractmethod
    def get_channel_selection_path(self) -> Optional[Path]:
        pass



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Callable, Any
from allencell_ml_segmenter.main.segmenter_layer import (
    ShapesLayer,
    ImageLayer,
    LabelsLayer,
)
import numpy as np
from napari.layers import Layer, Labels  # type: ignore
from napari.utils.events import Event as NapariEvent  # type: ignore


class IViewer(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def add_image(self, image: np.ndarray, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def get_image(self, name: str) -> Optional[ImageLayer]:
        pass

    @abstractmethod
    def get_all_images(self) -> list[ImageLayer]:
        pass

    @abstractmethod
    def add_shapes(self, name: str, face_color: str, mode: str) -> None:
        pass

    @abstractmethod
    def get_shapes(self, name: str) -> Optional[ShapesLayer]:
        pass

    @abstractmethod
    def get_all_shapes(self) -> list[ShapesLayer]:
        pass

    @abstractmethod
    def add_labels(self, data: np.ndarray, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def get_labels(self, name: str) -> Optional[LabelsLayer]:
        pass

    @abstractmethod
    def get_all_labels(self) -> list[LabelsLayer]:
        pass

    @abstractmethod
    def clear_layers(self) -> None:
        pass

    @abstractmethod
    def remove_layer(self, name: str) -> bool:
        pass

    @abstractmethod
    def contains_layer(self, name: str) -> bool:
        pass

    # TODO: refactor prediction/file_input_widget.py to not use this and use get_all_images instead
    @abstractmethod
    def get_layers(self) -> list[Layer]:
        pass

    @abstractmethod
    def subscribe_layers_change_event(
        self, function: Callable[[NapariEvent], None]
    ) -> None:
        pass

    @abstractmethod
    def get_seg_layers(self) -> list[Layer]:
        """
        Get only segmentation layers (which should be probability mappings) from the viewer.
        These are the layers that start with [seg].
        """
        pass

    @abstractmethod
    def insert_binary_map_into_layer(
        self, layer_name: str, img: np.ndarray, seg_layers: bool = False
    ) -> None:
        """
        Insert a thresholded image into the viewer.
        If a layer for this thresholded image already exists, the new image will replace the old one and refresh the viewer.
        If the layer does not exist, it will be added to the viewer in the correct place (on top of the original segmentation image:
        index_of_segmentation + 1 in the LayerList)
        """
        pass

    @abstractmethod
    def get_layers_nonthreshold(self) -> list[Layer]:
        """
        Get only layers which are not segmentation layers from the viewer.
        These are the layers that do not start with [threshold].
        """
        pass

    @abstractmethod
    def get_source_path(self, layer: Layer) -> Optional[Path]:
        pass

    @abstractmethod
    def get_all_layers_containing_prob_map(self) -> list[Layer]:
        pass

    @abstractmethod
    def clear_binary_map_from_layer(self, layer: Layer) -> None:
        pass



from allencell_ml_segmenter.curation.stacked_spinner import StackedSpinner
from allencell_ml_segmenter.widgets.input_button_widget import (
    InputButton,
    FileInputMode,
)
from allencell_ml_segmenter.widgets.label_with_hint_widget import LabelWithHint
from allencell_ml_segmenter.curation.curation_model import (
    CurationModel,
    CurationView,
    ImageType,
)
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QSizePolicy,
    QLabel,
    QFrame,
    QGridLayout,
    QVBoxLayout,
    QComboBox,
    QPushButton,
    QWidget,
)
from pathlib import Path
from napari.utils.notifications import show_info  # type: ignore
from typing import Optional


class CurationInputView(QWidget):
    """
    View for Curation UI
    """

    def __init__(self, curation_model: CurationModel) -> None:
        super().__init__()
        self._curation_model: CurationModel = curation_model
        layout: QVBoxLayout = QVBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum
        )

        self._title: QLabel = QLabel("CURATION UI", self)
        self._title.setObjectName("title")
        layout.addWidget(
            self._title, alignment=Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop  # type: ignore
        )

        frame: QFrame = QFrame()
        frame_layout: QVBoxLayout = QVBoxLayout()
        frame.setLayout(frame_layout)
        # uncomment to make frame visible
        # frame.setObjectName("frame")
        layout.addWidget(frame)

        input_images_label: QLabel = QLabel("Input images")
        frame_layout.addWidget(input_images_label)
        frame_layout.setSpacing(30)

        raw_grid_layout: QGridLayout = QGridLayout()
        raw_grid_layout.setVerticalSpacing(10)

        # First Row in Gridlayout
        raw_image_label: LabelWithHint = LabelWithHint("Raw")
        raw_image_label.set_hint(
            "Original microscopy images (.czi, .ome.tiff, .tiff)"
        )
        raw_grid_layout.addWidget(
            raw_image_label, 0, 0, alignment=Qt.AlignmentFlag.AlignLeft
        )
        raw_grid_layout.addWidget(
            QLabel("Directory"), 0, 1, alignment=Qt.AlignmentFlag.AlignRight
        )
        self.raw_directory_select: InputButton = InputButton(
            self._curation_model,
            self._on_raw_dir_select,
            "Select directory...",
            FileInputMode.DIRECTORY,
        )
        self.raw_dir_stacked_spinner = StackedSpinner(
            self.raw_directory_select
        )
        raw_grid_layout.addWidget(
            self.raw_dir_stacked_spinner,
            0,
            2,
            alignment=Qt.AlignmentFlag.AlignRight,
        )

        # Second Row in Gridlayout
        raw_grid_layout.addWidget(
            QLabel("Image channel"),
            1,
            1,
            alignment=Qt.AlignmentFlag.AlignRight,
        )
        self.raw_image_channel_combo: QComboBox = QComboBox()
        self.raw_image_channel_combo.currentIndexChanged.connect(
            self.raw_channel_selected
        )
        raw_grid_layout.addWidget(
            self.raw_image_channel_combo,
            1,
            2,
            alignment=Qt.AlignmentFlag.AlignLeft,
        )

        # add grid to frame
        frame_layout.addLayout(raw_grid_layout)

        seg1_grid_layout: QGridLayout = QGridLayout()
        seg1_grid_layout.setVerticalSpacing(10)

        # First Row in Gridlayout
        seg1_image_label: LabelWithHint = LabelWithHint("Seg 1")
        seg1_image_label.set_hint(
            "Segmentation of the structure of interest (.czi, .ome.tiff, .tiff)"
        )
        seg1_grid_layout.addWidget(
            seg1_image_label, 0, 0, alignment=Qt.AlignmentFlag.AlignLeft
        )
        seg1_grid_layout.addWidget(
            QLabel("Directory"), 0, 1, alignment=Qt.AlignmentFlag.AlignRight
        )
        # TODO update model accordingly
        self.seg1_directory_select: InputButton = InputButton(
            self._curation_model,
            self._on_seg1_dir_select,
            "Select directory...",
            FileInputMode.DIRECTORY,
        )
        self.seg1_dir_stacked_spinner = StackedSpinner(
            self.seg1_directory_select
        )
        seg1_grid_layout.addWidget(
            self.seg1_dir_stacked_spinner,
            0,
            2,
            alignment=Qt.AlignmentFlag.AlignRight,
        )

        # Second Row in Gridlayout
        seg1_grid_layout.addWidget(
            QLabel("Image channel"),
            1,
            1,
            alignment=Qt.AlignmentFlag.AlignRight,
        )
        self.seg1_image_channel_combo: QComboBox = QComboBox()
        self.seg1_image_channel_combo.currentIndexChanged.connect(
            self.seg1_channel_selected
        )
        seg1_grid_layout.addWidget(
            self.seg1_image_channel_combo,
            1,
            2,
            alignment=Qt.AlignmentFlag.AlignLeft,
        )

        # add grid to frame
        frame_layout.addLayout(seg1_grid_layout)

        seg2_grid_layout: QGridLayout = QGridLayout()
        seg2_grid_layout.setVerticalSpacing(10)

        # First Row in Gridlayout
        seg2_image_label: LabelWithHint = LabelWithHint("Seg 2 (OPTIONAL)")
        seg2_image_label.set_hint(
            "(Optional) Complementary segmentation, useful if Seg 1 fails predictably (e.g. a segmentation that works during mitosis to supplement an interphase segmentation)"
        )
        seg2_grid_layout.addWidget(
            seg2_image_label, 0, 0, alignment=Qt.AlignmentFlag.AlignLeft
        )
        seg2_grid_layout.addWidget(
            QLabel("Directory"), 0, 1, alignment=Qt.AlignmentFlag.AlignRight
        )
        # TODO update model accordingly
        self.seg2_directory_select: InputButton = InputButton(
            self._curation_model,
            self._on_seg2_dir_select,
            "Select directory...",
            FileInputMode.DIRECTORY,
        )
        self.seg2_dir_stacked_spinner = StackedSpinner(
            self.seg2_directory_select
        )
        seg2_grid_layout.addWidget(
            self.seg2_dir_stacked_spinner,
            0,
            2,
            alignment=Qt.AlignmentFlag.AlignRight,
        )

        # Second Row in Gridlayout
        seg2_grid_layout.addWidget(
            QLabel("Image channel"),
            1,
            1,
            alignment=Qt.AlignmentFlag.AlignRight,
        )
        self.seg2_image_channel_combo: QComboBox = QComboBox()
        self.seg2_image_channel_combo.currentIndexChanged.connect(
            self.seg2_channel_selected
        )
        seg2_grid_layout.addWidget(
            self.seg2_image_channel_combo,
            1,
            2,
            alignment=Qt.AlignmentFlag.AlignLeft,
        )

        # add grid to frame
        frame_layout.addLayout(seg2_grid_layout)

        self.start_btn: QPushButton = QPushButton("Start")
        self.start_btn.clicked.connect(self._on_start)
        frame_layout.addWidget(self.start_btn)

        # subscribers
        self._curation_model.channel_count_set.connect(self.update_channels)

    def _on_start(self) -> None:
        if any(
            [
                value is None
                for value in [
                    self._curation_model.get_image_directory(ImageType.RAW),
                    self._curation_model.get_selected_channel(ImageType.RAW),
                    self._curation_model.get_image_directory(ImageType.SEG1),
                    self._curation_model.get_selected_channel(ImageType.SEG1),
                ]
            ]
        ):
            show_info(
                "Please select a directory and channel for at least raw and seg1."
            )
            return

        if (
            self._curation_model.get_image_directory(ImageType.SEG2)
            is not None
            and self._curation_model.get_selected_channel(ImageType.SEG2)
            is None
        ):
            show_info("Please select a channel for seg2.")
            return

        self._curation_model.set_current_view(CurationView.MAIN_VIEW)
        self._curation_model.start_loading_images()

    def _set_to_loading(
        self, combobox: QComboBox, stacked_spinner: StackedSpinner
    ) -> None:
        stacked_spinner.start()
        combobox.clear()
        combobox.setPlaceholderText("loading channels...")
        combobox.setCurrentIndex(-1)
        combobox.setEnabled(False)

    def _set_to_stopped(
        self,
        combobox: QComboBox,
        stacked_spinner: StackedSpinner,
        input_button: InputButton,
    ) -> None:
        # reset everything to default
        stacked_spinner.stop()
        combobox.clear()
        combobox.setPlaceholderText("")
        combobox.setEnabled(True)
        input_button.clear_selection()

    def _on_raw_dir_select(self, dir: Path) -> None:
        self._set_to_loading(
            self.raw_image_channel_combo, self.raw_dir_stacked_spinner
        )
        self._curation_model.set_image_directory(ImageType.RAW, dir)

    def _on_seg1_dir_select(self, dir: Path) -> None:
        self._set_to_loading(
            self.seg1_image_channel_combo, self.seg1_dir_stacked_spinner
        )
        self._curation_model.set_image_directory(ImageType.SEG1, dir)

    def _on_seg2_dir_select(self, dir: Path) -> None:
        self._set_to_loading(
            self.seg2_image_channel_combo, self.seg2_dir_stacked_spinner
        )
        self._curation_model.set_image_directory(ImageType.SEG2, dir)

    def _populate_channel_combo(
        self, channel_combo: QComboBox, num_channels: int
    ) -> None:
        channel_combo.clear()
        if num_channels > 0:
            channel_combo.addItems([str(x) for x in range(num_channels)])
            channel_combo.setCurrentIndex(0)
            channel_combo.setEnabled(True)
        else:
            channel_combo.setPlaceholderText("")
            channel_combo.setEnabled(False)

    def update_channels(self, img_type: ImageType) -> None:
        if img_type == ImageType.RAW:
            self.update_raw_channels()
        elif img_type == ImageType.SEG1:
            self.update_seg1_channels()
        elif img_type == ImageType.SEG2:
            self.update_seg2_channels()
        else:
            raise RuntimeError("Unrecognized curation image type")

    def update_raw_channels(self) -> None:
        """
        Event handler when raw image directory is selected. Updates combobox to the correct number of channels in the
        images from the raw directory.
        """
        self.raw_dir_stacked_spinner.stop()
        channel_count: Optional[int] = self._curation_model.get_channel_count(
            ImageType.RAW
        )
        self._populate_channel_combo(
            self.raw_image_channel_combo,
            channel_count if channel_count is not None else 0,
        )
        self._curation_model.set_selected_channel(ImageType.RAW, 0)

    def update_seg1_channels(self) -> None:
        """
        Event handler when seg1 image directory is selected. Updates combobox to the correct number of channels in the
        images from the seg1 directory.
        """
        self.seg1_dir_stacked_spinner.stop()
        channel_count: Optional[int] = self._curation_model.get_channel_count(
            ImageType.SEG1
        )
        self._populate_channel_combo(
            self.seg1_image_channel_combo,
            channel_count if channel_count is not None else 0,
        )
        self._curation_model.set_selected_channel(ImageType.SEG1, 0)

    def update_seg2_channels(self) -> None:
        """
        Event handler when seg2 image directory is selected. Updates combobox to the correct number of channels in the
        images from the seg2 directory.
        """
        self.seg2_dir_stacked_spinner.stop()
        channel_count: Optional[int] = self._curation_model.get_channel_count(
            ImageType.SEG2
        )
        self._populate_channel_combo(
            self.seg2_image_channel_combo,
            channel_count if channel_count is not None else 0,
        )
        self._curation_model.set_selected_channel(ImageType.SEG2, 0)

    def raw_channel_selected(self, index: int) -> None:
        """
        Event handler when combobox channel selection is made. Sets the raw channel index in the model.
        """
        self._curation_model.set_selected_channel(ImageType.RAW, index)

    def seg1_channel_selected(self, index: int) -> None:
        """
        Event handler when combobox channel selection is made. Sets the seg1 channel index in the model.
        """
        self._curation_model.set_selected_channel(ImageType.SEG1, index)

    def seg2_channel_selected(self, index: int) -> None:
        """
        Event handler when combobox channel selection is made. Sets the seg2 channel index in the model.
        """
        self._curation_model.set_selected_channel(ImageType.SEG2, index)



import pytest
from pytestqt.qtbot import QtBot
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter._tests.fakes.fake_user_settings import (
    FakeUserSettings,
)
from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel
from allencell_ml_segmenter.main.main_model import MainModel

from allencell_ml_segmenter.training.model_selection_widget import (
    ModelSelectionWidget,
)
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
)


@pytest.fixture
def experiment_model() -> IExperimentsModel:
    return FakeExperimentsModel()


@pytest.fixture
def main_model() -> MainModel:
    """
    Fixture for MainModel testing.
    """
    return MainModel()


@pytest.fixture
def training_model(main_model: MainModel) -> TrainingModel:
    """
    Fixture that creates an instance of TrainingModel for testing.
    """
    return TrainingModel(  # instead of this, how about i mock os.listdir ?
        main_model, FakeExperimentsModel()
    )


@pytest.fixture
def model_selection_widget(
    main_model: MainModel,
    experiment_model: IExperimentsModel,
) -> ModelSelectionWidget:
    """
    Fixture that creates an instance of ModelSelectionWidget for testing.
    """
    return ModelSelectionWidget(
        main_model=main_model,
        experiments_model=experiment_model,
        user_settings=FakeUserSettings(),
    )


def test_radio_new_slot(
    qtbot: QtBot, model_selection_widget: ModelSelectionWidget
) -> None:
    """
    Test the slot connected to the "new model" radio button.
    """
    # ARRANGE
    model_selection_widget._combo_box_existing_models.setEnabled(True)

    # ACT (disable combo box)
    with qtbot.waitSignal(model_selection_widget._radio_new_model.toggled):
        model_selection_widget._radio_new_model.click()

    # ASSERT
    assert not model_selection_widget._combo_box_existing_models.isEnabled()


def test_radio_existing_slot_has_existing_experiments(
    qtbot: QtBot, model_selection_widget: ModelSelectionWidget
) -> None:
    setup_radio_existing_slot(qtbot, model_selection_widget)

    # ASSERT
    assert model_selection_widget._combo_box_existing_models.isEnabled()


def test_radio_existing_slot_has_no_existing_experiments(qtbot: QtBot) -> None:
    experiment_model_has_no_experiments: IExperimentsModel = (
        FakeExperimentsModel([])
    )
    model_selection_widget: ModelSelectionWidget = ModelSelectionWidget(
        main_model=MainModel(),
        experiments_model=experiment_model_has_no_experiments,
        user_settings=FakeUserSettings(),
    )
    setup_radio_existing_slot(qtbot, model_selection_widget)

    # ASSERT
    assert not model_selection_widget._combo_box_existing_models.isEnabled()


def setup_radio_existing_slot(
    qtbot: QtBot, model_selection_widget: ModelSelectionWidget
):
    """
    Test the slot connected to the "existing model" radio button.
    """

    # ACT (enable combo box)
    with qtbot.waitSignal(
        model_selection_widget._radio_existing_model.toggled
    ):
        model_selection_widget._radio_existing_model.click()


def test_select_existing_model_option(
    qtbot: QtBot,
    experiment_model: IExperimentsModel,
    model_selection_widget: ModelSelectionWidget,
) -> None:
    """
    Tests that the slots connected to the "start a new model" radio button and the existing model QCombBox properly set the model path field.
    """
    # ARRANGE - add arbitrary model path options to the QComboBox, since it does not come with default choices
    model_selection_widget._radio_new_model.setChecked(True)
    model_selection_widget._radio_existing_model.setChecked(False)
    with qtbot.waitSignal(
        model_selection_widget._radio_existing_model.toggled
    ):
        model_selection_widget._radio_existing_model.click()  # enables the combo box

    for i, experiment in enumerate(experiment_model.get_experiments()):
        # ACT - This is a compound action, consisting of selecting an option in the combo box and clicking the apply button
        # In prod, this could only be done for one option since the UI would then change.
        # As an isolated unit, we can test all options.
        # Invariant: options in existing_models combo were added in the order the appear in the model.
        model_selection_widget._combo_box_existing_models.setCurrentIndex(i)
        model_selection_widget._apply_btn.click()

        # ASSERT
        assert experiment_model.get_experiment_name() == experiment


def test_apply_button_enabled(
    model_selection_widget: ModelSelectionWidget,
    experiment_model: IExperimentsModel,
) -> None:
    """
    Test that the apply button is enabled when a model is selected.
    """
    # ARRANGE
    model_selection_widget._radio_existing_model.setChecked(True)

    # SANITY CHECK
    assert not model_selection_widget._apply_btn.isEnabled()

    # ACT
    model_selection_widget.select_experiment_name("dummy_experiment")

    # ASSERT
    assert model_selection_widget._apply_btn.isEnabled()


def test_text_input_enables_apply_button(
    model_selection_widget: ModelSelectionWidget,
) -> None:
    """
    Test that the apply button is disabled when a model is not selected.
    """
    # SANITY CHECK
    assert not model_selection_widget._apply_btn.isEnabled()

    # ACT
    model_selection_widget._experiment_name_input.setText("dummy_experiment")

    # ASSERT
    assert model_selection_widget._apply_btn.isEnabled()


def test_text_input_cleared_disables_apply_button(
    model_selection_widget: ModelSelectionWidget,
) -> None:
    """
    Test that the apply button is disabled when a typed model name is cleared.
    """
    # SANITY CHECK start in a state as if the user had endtered a new experiment name
    assert not model_selection_widget._apply_btn.isEnabled()
    model_selection_widget._experiment_name_input.setText("dummy_experiment")
    assert model_selection_widget._apply_btn.isEnabled()

    # ACT simulate clearing the text input
    model_selection_widget._experiment_name_input.setText("")

    # ASSERT UI behavior should revery to being as if the text input was never filled
    assert not model_selection_widget._apply_btn.isEnabled()


def test_combo_input_enables_apply_button_new_radio_disables(
    model_selection_widget: ModelSelectionWidget,
    qtbot: QtBot,
) -> None:
    """
    Test that the apply button Reacts to a model being selevted then deselected.
    """
    # ARRANGE
    model_selection_widget._radio_existing_model.setChecked(True)

    # SANITY CHECK (Initially no model is selected, so the apply button should NOT be enabled)
    assert not model_selection_widget._apply_btn.isEnabled()

    # ACT - select a model
    model_selection_widget._combo_box_existing_models.setCurrentIndex(1)

    # ASSERT - apply button SHOULD be enabled
    assert model_selection_widget._apply_btn.isEnabled()

    # ACT - select the "start a new model" radio button, clearing the model selection
    with qtbot.waitSignal(model_selection_widget._radio_new_model.toggled):
        model_selection_widget._radio_new_model.click()  # enables the combo box

    # ASSERT - apply button should NOT be enabled
    assert not model_selection_widget._apply_btn.isEnabled()


def test_new_model_enables_apply_button_new_radio_disables(
    model_selection_widget: ModelSelectionWidget,
    qtbot: QtBot,
) -> None:
    """
    Test that the apply button Reacts to a model being selevted then deselected.
    """
    # ARRANGE
    model_selection_widget._radio_new_model.setChecked(True)

    # SANITY CHECK Initially no model is selected, so the apply button should NOT be enabled
    assert not model_selection_widget._apply_btn.isEnabled()

    # ACT - select a model
    model_selection_widget._experiment_name_input.setText("dummy_experiment")

    # ASSERT - apply button SHOULD be enabled
    assert model_selection_widget._apply_btn.isEnabled()

    # ACT - select the "start a new model" radio button, clearing the model selection
    with qtbot.waitSignal(
        model_selection_widget._radio_existing_model.toggled
    ):
        model_selection_widget._radio_existing_model.click()  # enables the combo box

    # ASSERT - apply button should NOT be enabled
    assert not model_selection_widget._apply_btn.isEnabled()


def test_click_apply_btn(
    model_selection_widget: ModelSelectionWidget,
    experiment_model: IExperimentsModel,
) -> None:
    """
    Test that the apply button updates model.
    """
    # ARRANGE
    model_selection_widget.select_experiment_name("dummy_experiment")

    # Sanity check
    assert experiment_model.get_experiment_name() is None

    # ACT
    model_selection_widget._apply_btn.click()

    # ASSERT
    assert experiment_model.get_experiment_name() == "dummy_experiment"


# Disabled while I figure out what is wrong with this test
# def test_existing_experiment_apply(
#     model_selection_widget: ModelSelectionWidget,
#     experiment_model: IExperimentsModel,
# ) -> None:
#     """
#     Test that the apply button updates model.
#     """
#     # ACT
#     model_selection_widget._combo_box_existing_models.setIndex(1)

#     # ASSERT note that the model name is selected but not applied until the apply button is clicked
#     assert (
#         experiment_model.get_experiment_name_selection() == "0_exp"
#     )
#     assert experiment_model.get_experiment_name() is None

#     # ACT
#     model_selection_widget._apply_btn.click()

#     # ASSERT
#     assert experiment_model.get_experiment_name() == "dummy_experiment"


def test_new_experiment_apply(
    model_selection_widget: ModelSelectionWidget,
    experiment_model: IExperimentsModel,
) -> None:
    """
    Test that the apply button updates model.
    """
    # ACT
    model_selection_widget._experiment_name_input.setText("dummy_experiment")

    # ASSERT note that the model name is selected but not applied until the apply button is clicked
    assert experiment_model.get_experiment_name() is None

    # ACT
    model_selection_widget._apply_btn.click()

    # ASSERT
    assert experiment_model.get_experiment_name() == "dummy_experiment"



from pathlib import Path
from typing import Optional
from allencell_ml_segmenter.config.i_user_settings import IUserSettings
from allencell_ml_segmenter.utils.experiment_utils import ExperimentUtils

import copy

from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel


class ExperimentsModel(IExperimentsModel):
    def __init__(self, config: IUserSettings) -> None:
        super().__init__()
        self.user_settings: IUserSettings = config

        # options
        self.experiments: list[str] = []
        self.refresh_experiments()

    def refresh_experiments(self) -> None:
        # TODO: make a FileUtils method for this?
        self.experiments = []
        user_exp_path: Optional[Path] = (
            self.user_settings.get_user_experiments_path()
        )
        if user_exp_path is not None:
            for experiment in user_exp_path.iterdir():
                if (
                    experiment.is_dir()
                    and self._is_cyto_dl_experiment(experiment)
                    and experiment not in self.experiments
                    and not experiment.name.startswith(".")
                ):
                    self.experiments.append(experiment.name)
            self.experiments.sort()

    """
    Returns a defensive copy of Experiments list.
    """

    def _is_cyto_dl_experiment(self, experiment: Path) -> bool:
        # Heuristic for checking if dir is a cyto-dl experiment
        csv_path: Path = experiment / "data" / "train.csv"
        checkpoints_path: Path = experiment / "checkpoints"
        return checkpoints_path.exists() or csv_path.exists()

    def get_experiments(self) -> list[str]:
        return copy.deepcopy(self.experiments)

    def get_user_settings(self) -> IUserSettings:
        return self.user_settings

    def get_user_experiments_path(self) -> Optional[Path]:
        return self.get_user_settings().get_user_experiments_path()

    # TODO: possibly refactor to get rid of exp name and checkpoint params?
    def get_model_checkpoints_path(
        self, experiment_name: Optional[str], checkpoint: Optional[str]
    ) -> Path:
        """
        Gets checkpoints for model path
        """
        if experiment_name is None:
            raise ValueError(
                "Experiment name cannot be None in order to get model_checkpoint_path"
            )

        if checkpoint is None:
            raise ValueError(
                "Checkpoint cannot be None in order to get model_checkpoint_path"
            )

        user_exp_path: Optional[Path] = self.get_user_experiments_path()
        if user_exp_path is None:
            raise ValueError("User experiments path cannot be None")

        return user_exp_path / experiment_name / "checkpoints" / checkpoint

    def _get_exp_path(self) -> Optional[Path]:
        user_exp_path: Optional[Path] = self.get_user_experiments_path()
        exp_name: Optional[str] = self.get_experiment_name()
        if exp_name is not None and user_exp_path is not None:
            return user_exp_path / exp_name
        return None

    def get_csv_path(self) -> Optional[Path]:
        exp_path: Optional[Path] = self._get_exp_path()
        if exp_path is not None:
            return exp_path / "data"
        return None

    def get_metrics_csv_path(self) -> Optional[Path]:
        exp_path: Optional[Path] = self._get_exp_path()
        if exp_path is not None:
            return exp_path / "csv"
        return None

    def get_cache_dir(self) -> Optional[Path]:
        exp_path: Optional[Path] = self._get_exp_path()
        if exp_path is not None:
            return exp_path / "cache"
        return None

    def get_latest_metrics_csv_version(self) -> int:
        """
        Returns version number of the most recent version directory within
        the cyto-dl CSV folder (self._csv_path) or -1 if no version directories
        exist
        """
        last_version: int = -1
        csv_path: Optional[Path] = self.get_metrics_csv_path()
        if csv_path is not None and csv_path.exists():
            for child in csv_path.glob("version_*"):
                if child.is_dir():
                    version_str: str = child.name.split("_")[-1]
                    try:
                        last_version = (
                            int(version_str)
                            if int(version_str) > last_version
                            else last_version
                        )
                    except ValueError:
                        continue
        return last_version

    def get_latest_metrics_csv_path(self) -> Optional[Path]:
        version: int = self.get_latest_metrics_csv_version()
        csv_path: Optional[Path] = self.get_metrics_csv_path()
        return (
            csv_path / f"version_{version}" / "metrics.csv"
            if version >= 0 and csv_path is not None
            else None
        )

    def get_train_config_path(
        self, experiment_name: Optional[str] = None
    ) -> Path:
        if experiment_name is not None:
            # user is getting a config for an existing experiment
            user_exp_path: Optional[Path] = self.get_user_experiments_path()
            if user_exp_path is None:
                raise ValueError(
                    "user_exp_path cannot be None if experiment_name is also None in get_train_config_path"
                )
            return user_exp_path / experiment_name / "train_config.yaml"
        else:
            user_exp_path = self._get_exp_path()
            # get config for currently selected experiment
            if user_exp_path is None:
                raise ValueError(
                    "user_exp_path cannot be None if experiment_name is also None in get_train_config_path"
                )
            return user_exp_path / "train_config.yaml"

    def get_current_epoch(self) -> Optional[int]:
        ckpt: Optional[Path] = self.get_best_ckpt()
        if not ckpt:
            return None
        # assumes checkpoint format: path/to/checkpoint/epoch_001.ckpt

        return int(ckpt.name.split(".")[0].split("_")[-1])

    def get_best_ckpt(self) -> Optional[Path]:
        user_exp_path: Optional[Path] = (
            self.user_settings.get_user_experiments_path()
        )
        if user_exp_path is None:
            # need user_exp_path in order to get experiments list
            return None

        selected_experiment: Optional[str] = self.get_experiment_name()
        if selected_experiment is None:
            return None
        return ExperimentUtils.get_best_ckpt(
            user_exp_path, selected_experiment
        )

    def get_channel_selection_path(self) -> Optional[Path]:
        csv_path: Optional[Path] = self.get_csv_path()
        if csv_path is None:
            return None
        return csv_path / "selected_channels.json"



from collections import OrderedDict
from pathlib import Path
from typing import Callable, Optional
from bioio import BioImage
from bioio.writers import OmeTiffWriter
from napari.layers import Layer, Labels  # type: ignore
import numpy as np
from napari.utils.notifications import show_info  # type: ignore

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.thresholding.thresholding_model import (
    ThresholdingModel,
)
from allencell_ml_segmenter.core.task_executor import (
    NapariThreadTaskExecutor,
    ITaskExecutor,
)
from allencell_ml_segmenter.main.viewer import IViewer
from cyto_dl.models.im2im.utils.postprocessing.auto_thresh import AutoThreshold  # type: ignore


class ThresholdingService(Subscriber):
    def __init__(
        self,
        thresholding_model: ThresholdingModel,
        experiments_model: ExperimentsModel,
        main_model: MainModel,
        viewer: IViewer,
        task_executor: ITaskExecutor = NapariThreadTaskExecutor.global_instance(),
    ):
        super().__init__()
        # Models
        self._thresholding_model: ThresholdingModel = thresholding_model
        self._experiments_model: ExperimentsModel = experiments_model
        self._main_model: MainModel = main_model

        # napari viewer
        self._viewer: IViewer = viewer

        # Task Executor
        self._task_executor: ITaskExecutor = task_executor

        self._thresholding_model.subscribe(
            Event.ACTION_EXECUTE_THRESHOLDING,
            self,
            self._on_threshold_changed,
        )

        self._thresholding_model.subscribe(
            Event.ACTION_SAVE_THRESHOLDING_IMAGES,
            self,
            self._save_thresholded_images,
        )

        self._thresholding_model.subscribe(
            Event.ACTION_THRESHOLDING_DISABLED,
            self,
            self._remove_all_binary_maps,
        )

    def _handle_thresholding_error(self, error: Exception) -> None:
        show_info("Thresholding failed: " + str(error))

    def _on_threshold_changed(self, _: Event) -> None:
        # Check to see if user has selected a thresholding method
        if (
            self._thresholding_model.is_threshold_enabled()
            or self._thresholding_model.is_autothresholding_enabled()
        ):
            # get all layers with a prob map
            layers_containing_prob_map: list[Layer] = (
                self._thresholding_model.get_thresholding_layers()
            )

            # determine thresholding function to use based on user selection
            if self._thresholding_model.is_autothresholding_enabled():
                thresh_function: Callable = AutoThreshold(
                    self._thresholding_model.get_autothresholding_method()
                )
            elif self._thresholding_model.is_threshold_enabled():
                thresh_function = self._threshold_image

            # selected layers in the ui
            selected_idx: list[int] = (
                self._thresholding_model.get_selected_idx()
            )

            for idx, layer in enumerate(layers_containing_prob_map):
                # for selected layers, perform thresholding and display the result in the viewer
                if idx in selected_idx:
                    # Creating helper functions for mypy strict typing
                    # Thresholding function
                    def thresholding_task(
                        layer_instance: Layer = layer,
                    ) -> np.ndarray:
                        # INVARIANT: a segmentation layer must have prob_map in its metadata if it came from our plugin
                        # so we are only supporting thresholding images that are from the plugin itself.
                        if (
                            not isinstance(layer_instance.metadata, dict)
                            or "prob_map" not in layer_instance.metadata
                        ):
                            raise ValueError(
                                "Layer metadata must be a dictionary containing the 'prob_map' key in order to threshold."
                            )
                        # This thresholding task returns a binary map
                        return thresh_function(
                            layer_instance.metadata["prob_map"]
                        )

                    # On return, display the binary map that was produced from thresholding
                    def on_return(
                        resulting_binary_map: np.ndarray,
                        layer_instance: Layer = layer,
                    ) -> None:
                        self._viewer.insert_binary_map_into_layer(
                            layer_instance,
                            resulting_binary_map,
                            self._main_model.are_predictions_in_viewer(),
                        )

                    # Task executor to handle this thresholding task
                    self._task_executor.exec(
                        task=thresholding_task,
                        # lambda functions capture variables by reference so need to pass layer as a default argument
                        on_return=on_return,
                        on_error=self._handle_thresholding_error,
                    )
                else:
                    # If not selected (or unselected)- clear the binary map from the napari viewer.
                    self._viewer.clear_binary_map_from_layer(layer)

    def _save_thresholded_images(self, _: Event) -> None:
        images_to_threshold: list[Path] = (
            self._thresholding_model.get_input_files_as_list()
        )
        if self._thresholding_model.is_autothresholding_enabled():
            thresh_function: Callable = AutoThreshold(
                self._thresholding_model.get_autothresholding_method()
            )
        else:
            thresh_function = self._threshold_image
        for path in images_to_threshold:
            image = BioImage(path)
            try:
                self._save_thresh_image(thresh_function(image.data), path.name)
            except Exception as e:
                self._handle_thresholding_error(e)

    def _save_thresh_image(
        self, image: np.ndarray, original_image_name: str
    ) -> None:
        output_directory: Optional[Path] = (
            self._thresholding_model.get_output_directory()
        )
        if output_directory is not None:
            new_image_path: Path = (
                output_directory / f"threshold_{original_image_name}"
            )
            OmeTiffWriter.save(image, str(new_image_path))

    def _threshold_image(self, image: np.ndarray) -> np.ndarray:
        threshold_value: int = (
            self._thresholding_model.get_thresholding_value()
        )
        return (image > threshold_value).astype(np.uint8)

    def _remove_all_binary_maps(self, _: Event) -> None:
        for layer in self._thresholding_model.get_thresholding_layers():
            self._viewer.clear_binary_map_from_layer(layer)



from pathlib import Path

import pytest
from pytestqt.qtbot import QtBot

import allencell_ml_segmenter
from allencell_ml_segmenter._tests.fakes.fake_viewer import FakeViewer
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.prediction.model import (
    PredictionModel,
)
from allencell_ml_segmenter.core.image_data_extractor import (
    FakeImageDataExtractor,
)
from allencell_ml_segmenter.core.file_input_model import (
    InputMode,
    FileInputModel,
)
from allencell_ml_segmenter.prediction.view import PredictionView


@pytest.fixture
def main_model() -> MainModel:
    """
    Returns a MainModel instance for testing.
    """
    return MainModel()


@pytest.fixture
def prediction_view(main_model: MainModel, qtbot: QtBot) -> PredictionView:
    """
    Returns a PredictionView instance for testing.
    """
    prediction_model: PredictionModel = PredictionModel()
    return PredictionView(
        main_model, prediction_model, FileInputModel(), FakeViewer()
    )


def test_prediction_view(
    prediction_view: PredictionView, main_model: MainModel
) -> None:
    """
    Tests that the PredictionView correctly sets the current view to itself
    """
    # ACT
    main_model.dispatch(Event.PROCESS_TRAINING_COMPLETE)

    # ASSERT
    assert main_model.get_current_view() == prediction_view


def test_show_results(main_model: MainModel) -> None:
    """
    Testing the showresults that runs after a prediction run
    """
    # ARRANGE
    file_input_model: FileInputModel = FileInputModel()
    prediction_model: PredictionModel = PredictionModel()
    file_input_model.set_output_directory(
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "output_test_folder"
    )
    file_input_model.set_input_mode(InputMode.FROM_NAPARI_LAYERS)
    file_input_model.set_selected_paths(
        [Path("output_1.tiff"), Path("output_2.tiff")]
    )
    file_input_model.set_image_input_channel_index(0)
    fake_viewer: FakeViewer = FakeViewer()

    prediction_view: PredictionView = PredictionView(
        main_model,
        prediction_model,
        file_input_model,
        fake_viewer,
        img_data_extractor=FakeImageDataExtractor.global_instance(),
    )

    # ACT
    prediction_view.showResults()

    # ASSERT
    assert len(fake_viewer.get_all_labels()) == 2
    assert len(fake_viewer.get_all_images()) == 2
    assert fake_viewer.contains_layer("[raw] output_1.tiff")
    assert fake_viewer.contains_layer("[seg] output_1.tiff")
    assert fake_viewer.contains_layer("[raw] output_2.tiff")
    assert fake_viewer.contains_layer("[seg] output_2.tiff")


def test_show_results_non_empty_folder(main_model: MainModel) -> None:
    """
    Testing that only the new images in a folder will be shown after prediction.
    """
    # ARRANGE
    file_input_model: FileInputModel = FileInputModel()
    prediction_model: PredictionModel = PredictionModel()
    file_input_model.set_output_directory(
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "output_test_folder_extra"
    )
    file_input_model.set_input_mode(InputMode.FROM_NAPARI_LAYERS)
    file_input_model.set_selected_paths(
        [Path("output_3.tiff"), Path("output_4.tiff")]
    )
    file_input_model.set_image_input_channel_index(0)
    fake_viewer: FakeViewer = FakeViewer()

    prediction_view: PredictionView = PredictionView(
        main_model,
        prediction_model,
        file_input_model,
        fake_viewer,
        img_data_extractor=FakeImageDataExtractor.global_instance(),
    )

    # ACT
    prediction_view.showResults()

    # ASSERT
    assert len(fake_viewer.get_all_labels()) == 2
    assert len(fake_viewer.get_all_images()) == 2
    assert fake_viewer.contains_layer("[raw] output_3.tiff")
    assert fake_viewer.contains_layer("[seg] output_3.tiff")
    assert fake_viewer.contains_layer("[raw] output_4.tiff")
    assert fake_viewer.contains_layer("[seg] output_4.tiff")



import csv
import platform

from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.prediction.model import (
    PredictionModel,
)
from allencell_ml_segmenter.core.file_input_model import (
    InputMode,
    FileInputModel,
)

from allencell_ml_segmenter.utils.cuda_util import CUDAUtils
from allencell_ml_segmenter.utils.file_utils import FileUtils

from pathlib import Path
from typing import Union, Any, List, Optional

from cyto_dl.api.model import CytoDLModel  # type: ignore
from napari.utils.notifications import show_warning  # type: ignore


class PredictionService(Subscriber):
    """
    Interface for training a model or predicting using a model.
    Uses cyto-dl for model training and inference.
    #TODO create an ABC for cyto-service and have prediction_service inherit it
    """

    def __init__(
        self,
        prediction_model: PredictionModel,
        file_input_model: FileInputModel,
        experiments_model: ExperimentsModel,
    ):
        super().__init__()
        self._prediction_model: PredictionModel = prediction_model
        self._file_input_model: FileInputModel = file_input_model
        self._experiments_model: ExperimentsModel = experiments_model

        self._prediction_model.subscribe(
            Event.PROCESS_PREDICTION,
            self,
            self._predict_model,
        )

        self._prediction_model.subscribe(
            Event.ACTION_PREDICTION_SETUP,
            self,
            self._prediction_setup,
        )

    def _predict_model(self, _: Event) -> None:
        """
        Predict segmentations using model according to spec
        """
        cyto_api: CytoDLModel = CytoDLModel()
        cyto_api.load_config_from_file(
            self._experiments_model.get_train_config_path()
        )
        ckpt: Optional[Path] = self._experiments_model.get_best_ckpt()
        if ckpt is None:
            raise RuntimeError("No checkpoint. Cannot predict")
        cyto_api.override_config(self.build_overrides(ckpt))
        cyto_api.predict()

    def _prediction_setup(self, _: Event) -> None:
        if self._able_to_continue_prediction():
            self._write_csv_for_prediction()

    def _able_to_continue_prediction(self) -> bool:
        # Check to see if experiment selected
        experiment_name: Optional[str] = (
            self._experiments_model.get_experiment_name()
        )
        if experiment_name is None:
            show_warning(
                "Please select an experiment before running prediction."
            )
            return False

        # Check to see if training has occurred with the selected experiment.
        training_config: Path = self._experiments_model.get_train_config_path()
        if not training_config.exists():
            show_warning(
                f"Please train with the experiment: {experiment_name} before running a prediction."
            )
            return False

        # Check to see the user has specified a ckpt to use.
        if self._experiments_model.get_best_ckpt() is None:
            show_warning(
                f"Please select a checkpoint to run predictions with."
            )
            return False

        # Check to see the user has specified an output folder to use.
        if self._file_input_model.get_output_directory() is None:
            show_warning(
                f"Please select an output folder to save predictions to."
            )
            return False
        return True

    def _write_csv_for_prediction(self) -> None:
        """
        If needed, write csv's for predictions, and set total number of images to be processed
        """
        # Check to see if user has selected an input mode
        input_mode_selected: Optional[InputMode] = (
            self._file_input_model.get_input_mode()
        )
        if not input_mode_selected:
            show_warning(
                "Please select input images before running prediction."
            )
            # dont set state if we have an error in setup
        elif input_mode_selected == InputMode.FROM_PATH:
            self._prediction_model.set_total_num_images(
                self._setup_inputs_from_path()
            )
        elif input_mode_selected == InputMode.FROM_NAPARI_LAYERS:
            self._prediction_model.set_total_num_images(
                self._setup_inputs_from_napari()
            )

    def build_overrides(self, checkpoint: Path) -> dict[str, Any]:
        """
        Build an overrides list for the cyto-dl API containing the
        overrides requried to run predictions, formatted as cyto-dl expects.
        """
        overrides: dict[str, Any] = dict()
        # Default overrides needed for prediction
        overrides["test"] = False
        overrides["train"] = False
        overrides["mode"] = "predict"
        overrides["task_name"] = "predict_task_from_app"
        # Need these overrides to load in csv's
        overrides["data.columns"] = ["raw", "split"]
        overrides["data.split_column"] = "split"
        overrides["checkpoint.ckpt_path"] = str(checkpoint)
        overrides["checkpoint.strict"] = False
        overrides["checkpoint.weights_only"] = True
        # This override is needed to for inference with cyto-dl, because we default to auto otherwise
        overrides["data.batch_size"] = 1

        input_path: Optional[Path] = (
            self._file_input_model.get_input_image_path()
        )
        if input_path is None:
            raise RuntimeError("Path to prediction input undefined")

        overrides["data.path"] = str(input_path)

        # overrides from model
        # if output_dir is not set, will default to saving in the experiment folder
        output_dir: Optional[Path] = (
            self._file_input_model.get_output_directory()
        )
        if output_dir:
            overrides["paths.output_dir"] = str(output_dir)

        # if channel is not set, will default to same channel used to train
        channel: Optional[int] = (
            self._file_input_model.get_image_input_channel_index()
        )
        if channel is not None:
            overrides["input_channel"] = channel

        # selecting hardware- GPU if available (and correct drivers installed),
        # CPU otherwise.
        if CUDAUtils.cuda_available():
            overrides["trainer.accelerator"] = "gpu"
        else:
            overrides["trainer.accelerator"] = "cpu"

        # pytorch needs num_workers = 0 for mac osx
        # keeps prediction on main thread
        if platform.system() == "Darwin":
            overrides["data.num_workers"] = 0

        return overrides

    def write_csv_for_inputs(self, list_images: List[Path]) -> None:
        """
        write csv for inputs and return the total number of images
        """
        data_folder: Optional[Path] = self._experiments_model.get_csv_path()
        if data_folder is not None:
            (data_folder / "prediction_csv").mkdir(parents=True, exist_ok=True)
            csv_path: Path = data_folder / "prediction_csv" / "test_csv.csv"
            with open(csv_path, "w") as file:
                writer = csv.writer(file)
                writer.writerow(["", "raw", "split"])
                for i, path_of_image in enumerate(list_images):
                    writer.writerow([str(i), str(path_of_image), "test"])

            self._file_input_model.set_input_image_path(csv_path)

    def _setup_inputs_from_path(self) -> int:
        """
        setup inputs from path and return total number of images to predict
        """
        # User has selected a directory or a csv as input images
        input_path: Optional[Path] = (
            self._file_input_model.get_input_image_path()
        )
        if input_path is not None and input_path.is_dir():
            all_files: list[Path] = (
                FileUtils.get_all_files_in_dir_ignore_hidden(input_path)
            )
            # if input path selected is a directory, we need to manually write a CSV for cyto-dl
            self.write_csv_for_inputs(all_files)
            return len(all_files)
        elif input_path is not None and input_path.suffix == ".csv":
            return self._grab_csv_data_rows(input_path)
        else:
            # This should not be possible with FileInputWidget- throw an error.
            raise ValueError(
                "Somehow the user has selected a non-csv/directory for input images. Should not be possible with FileInputWidget"
            )
        # if a csv is selected, do nothing

    def _grab_csv_data_rows(self, path: Path) -> int:
        file = open(path, "r+")
        reader = csv.reader(file)
        return len(list(reader)) - 1  # ignore header for rowcount

    def _setup_inputs_from_napari(self) -> Optional[int]:
        """
        setup inputs from napari layers and return total number of images to predict
        """
        # User has selected napari image layers as input images
        selected_paths_from_napari: Optional[list[Path]] = (
            self._file_input_model.get_selected_paths()
        )
        if (
            selected_paths_from_napari is None
            or len(selected_paths_from_napari) < 1
        ):
            # No image layers selected
            show_warning(
                "Please select at least 1 image from the napari layer before running prediction."
            )
            return None
        else:
            # If user selects input images from napari, we need to manually write a csv for cyto-dl
            self.write_csv_for_inputs(selected_paths_from_napari)
            return len(selected_paths_from_napari)



from typing import Optional
import webbrowser
from qtpy.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QSizePolicy,
    QFrame,
    QGridLayout,
    QComboBox,
    QRadioButton,
    QLineEdit,
    QPushButton,
    QStackedWidget,
    QLabel,
)
from qtpy.QtCore import Qt

from allencell_ml_segmenter.config.i_user_settings import IUserSettings
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.utils.s3.s3_bucket_constants import (
    ENABLE_MODEL_DOWNLOADS,
)
from allencell_ml_segmenter.widgets.label_with_hint_widget import LabelWithHint
from allencell_ml_segmenter.widgets.model_download_dialog import (
    ModelDownloadDialog,
)


class ModelSelectionWidget(QWidget):
    """
    A widget for segmentation model selection.
    """

    TITLE_TEXT: str = "Segmentation model"
    USER_GUIDE_TEXT: str = "User Guide"
    GITHUB_TEXT: str = "GitHub"
    FORUM_TEXT: str = "Forum"
    WEBSITE_TEXT: str = "Website"
    EXPERIMENTS_HOME_TEXT: str = "Experiments Home"
    DOWNLOAD_EXPERIMENTS_TEXT: str = "Download Models"

    def __init__(
        self,
        main_model: MainModel,
        experiments_model: IExperimentsModel,
        user_settings: IUserSettings,
    ) -> None:
        super().__init__()

        self._main_model: MainModel = main_model
        self._experiments_model: IExperimentsModel = experiments_model
        self._user_settings: IUserSettings = user_settings

        # widget skeleton
        layout: QVBoxLayout = QVBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum
        )

        self._title: LabelWithHint = LabelWithHint(
            ModelSelectionWidget.TITLE_TEXT,
        )
        self._title.set_hint(
            "A type of ML model that separates the structures of interest from their background in a 2D/3D mircroscopy image"
        )
        self._title.setStyleSheet("padding-top: 12px")

        self._model_name_label: QLabel = QLabel()
        self._model_name_label.setAlignment(Qt.AlignmentFlag.AlignLeft)

        # Help menu, implemented using a combo box to conform to mockup
        self.help_combo_box: QComboBox = QComboBox()
        self.help_combo_box.setFixedWidth(100)
        # Add placeholder "Help" text idx 0
        self.help_combo_box.addItem("Help")
        self.help_combo_box.setCurrentIndex(0)
        self.help_combo_box.addItems(
            [
                ModelSelectionWidget.USER_GUIDE_TEXT,
                ModelSelectionWidget.GITHUB_TEXT,
                ModelSelectionWidget.FORUM_TEXT,
                ModelSelectionWidget.WEBSITE_TEXT,
                ModelSelectionWidget.EXPERIMENTS_HOME_TEXT,
                ModelSelectionWidget.DOWNLOAD_EXPERIMENTS_TEXT,
            ]
        )
        self.help_combo_box.currentTextChanged.connect(
            self._help_combo_handler
        )

        plugin_title_widget_layout: QHBoxLayout = QHBoxLayout()
        plugin_title_widget_layout.addWidget(
            QLabel("Allen Cell & Structure Segmenter - MACHINE LEARNING")
        )
        plugin_title_widget_layout.addWidget(
            self.help_combo_box, alignment=Qt.AlignmentFlag.AlignRight
        )

        layout.setContentsMargins(20, 20, 20, 20)
        layout.addLayout(plugin_title_widget_layout)

        # layout for model labels
        label_widget_layout: QHBoxLayout = QHBoxLayout()
        layout.addLayout(label_widget_layout)
        label_widget_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        label_widget_layout.setSpacing(0)
        label_widget_layout.addWidget(self._title)
        label_widget_layout.addWidget(
            self._model_name_label, alignment=Qt.AlignmentFlag.AlignLeft
        )
        label_widget_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        # TODO: hints for widget titles?

        frame: QFrame = QFrame()
        frame_layout: QVBoxLayout = QVBoxLayout()
        frame.setLayout(frame_layout)
        # uncomment to make frame visible
        # frame.setObjectName("frame")
        layout.addWidget(frame)

        # existing model selection components must be initialized before the new/existing model radios
        self._combo_box_existing_models: QComboBox = QComboBox()
        self._combo_box_existing_models.setCurrentIndex(-1)
        self._combo_box_existing_models.setEnabled(False)
        self._combo_box_existing_models.setMinimumWidth(306)
        self._refresh_experiment_options()
        self._combo_box_existing_models.currentTextChanged.connect(
            self._model_combo_handler
        )

        # model selection components
        top_grid_layout: QGridLayout = QGridLayout()

        self._radio_new_model: QRadioButton = QRadioButton()
        self._radio_new_model.toggled.connect(self._model_radio_handler)
        # initialize the radio button and combos / tabs to match the model state
        top_grid_layout.addWidget(self._radio_new_model, 0, 0)

        self._experiment_name_input: QLineEdit = QLineEdit()
        self._experiment_name_input.setPlaceholderText("Name your model")
        self._experiment_name_input.textChanged.connect(
            self._experiment_name_input_handler
        )

        label_new_model: LabelWithHint = LabelWithHint("Start a new model")
        label_new_model.set_hint(
            "Train a segmentation model from scratch or start from a previously trained model"
        )
        top_grid_layout.addWidget(label_new_model, 0, 1)
        top_grid_layout.addWidget(self._experiment_name_input, 0, 2)

        self._radio_existing_model: QRadioButton = QRadioButton()
        self._radio_existing_model.toggled.connect(self._model_radio_handler)
        # initialize the radio button and combos / tabs to match the model state
        top_grid_layout.addWidget(self._radio_existing_model, 1, 0)
        top_grid_layout.addWidget(
            LabelWithHint(
                label_text="Select an existing model",
                hint="Choose a previously trained model to segment to your data",
            ),
            1,
            1,
        )
        top_grid_layout.addWidget(self._combo_box_existing_models, 1, 2)

        self._apply_change_stacked_widget = QStackedWidget()

        apply_model_layout = QVBoxLayout()
        apply_model_layout.addLayout(top_grid_layout)

        apply_model_widget = QWidget()
        apply_model_widget.setLayout(apply_model_layout)
        self._apply_change_stacked_widget.addWidget(apply_model_widget)
        self._apply_btn: QPushButton = QPushButton("Apply")
        self._apply_btn.setEnabled(False)
        self._apply_btn.clicked.connect(self._handle_apply_model)
        apply_model_layout.addWidget(self._apply_btn)

        inner_frame_layout = QVBoxLayout()
        inner_frame_layout.addWidget(self._apply_change_stacked_widget)
        frame_layout.addLayout(inner_frame_layout)

        self._experiments_model.subscribe(
            Event.ACTION_REFRESH, self, self._handle_process_event
        )

    def _handle_apply_model(self) -> None:
        if self._experiment_name_selection is None:
            raise RuntimeError("Cannot apply an empty model name")

        self._experiments_model.apply_experiment_name(
            self._experiment_name_selection
        )
        self._title.set_value_text("    " + self._experiment_name_selection)
        self._apply_change_stacked_widget.setVisible(False)

    def _model_combo_handler(self, experiment_name: str) -> None:
        """
        Triggered when the user selects a model from the _combo_box_existing_models.
        Sets the model path in the model.
        """
        self.select_experiment_name(experiment_name)

    def _experiment_name_input_handler(self, text: str) -> None:
        """
        Triggered when the user types in the _experiment_name_input.
        Sets the model name in the model.
        """
        self.select_experiment_name(text)

    def _help_combo_handler(self, text: str) -> None:
        """
        Triggered when the user selects an option from the help combo box.
        Opens the selected help page.
        """
        if text == ModelSelectionWidget.USER_GUIDE_TEXT:
            webbrowser.open(
                "https://allencell.github.io/allencell-segmenter-ml/index.html"
            )
        elif text == ModelSelectionWidget.GITHUB_TEXT:
            webbrowser.open(
                "https://github.com/AllenCell/allencell-ml-segmenter"
            )
        elif text == ModelSelectionWidget.FORUM_TEXT:
            webbrowser.open("https://forum.image.sc/tag/segmenter")
        elif text == ModelSelectionWidget.WEBSITE_TEXT:
            webbrowser.open("https://www.allencell.org/segmenter.html")
        elif text == ModelSelectionWidget.EXPERIMENTS_HOME_TEXT:
            self._user_settings.display_change_user_experiments_home(
                parent=self
            )
            self._refresh_experiment_options()
        elif (
            text == ModelSelectionWidget.DOWNLOAD_EXPERIMENTS_TEXT
            and ENABLE_MODEL_DOWNLOADS
        ):
            dialog = ModelDownloadDialog(self, self._experiments_model)
            dialog.exec()
            # once all models are downloaded, one final refresh to load them into existing models dropdown
            self._refresh_experiment_options()

        # reset the combo box, so that it bahaves more like a menu
        self.help_combo_box.setCurrentIndex(0)

    def _model_radio_handler(self) -> None:
        self._main_model.set_new_model(self._radio_new_model.isChecked())
        self._combo_box_existing_models.setCurrentIndex(-1)
        self._experiment_name_input.clear()
        self.select_experiment_name(None)
        if self._radio_new_model.isChecked():
            """
            Triggered when the user selects the "start a new model" radio button.
            Enables and disables relevent controls.
            """
            self._combo_box_existing_models.setEnabled(False)
            self._experiment_name_input.setEnabled(True)

        if self._radio_existing_model.isChecked():
            """
            Triggered when the user selects the "existing model" radio button.
            Enables and disables relevent controls.
            """
            self._combo_box_existing_models.setEnabled(
                self._experiments_model.get_experiments() != []
            )
            self._experiment_name_input.setEnabled(False)

    def _handle_process_event(self, _: Optional[Event] = None) -> None:
        """
        Refreshes the experiments in the _combo_box_existing_models.
        """
        if self._radio_new_model.isChecked():
            self._refresh_experiment_options()

    def select_experiment_name(self, name: Optional[str]) -> None:
        """
        Sets experiment name
        """
        if name == "":
            self._experiment_name_selection = None
        else:
            self._experiment_name_selection = name

        experiment_selected = self._experiment_name_selection is not None
        self._apply_btn.setEnabled(experiment_selected)

    def _refresh_experiment_options(self) -> None:
        self._experiments_model.refresh_experiments()
        self._combo_box_existing_models.clear()
        self._combo_box_existing_models.addItems(
            self._experiments_model.get_experiments()
        )
        self._combo_box_existing_models.setCurrentIndex(-1)
        self._combo_box_existing_models.setPlaceholderText(
            "No existing models"
            if self._experiments_model.get_experiments() == []
            else "Select an existing model"
        )



from pathlib import Path
from unittest.mock import Mock, mock_open, patch, create_autospec

import allencell_ml_segmenter
from allencell_ml_segmenter.utils.zip_file import (
    ZipFileManager,
    IZipFileManager,
)


def test_write_to_file() -> None:
    # Arrange
    to_write: bytes = bytes("abcde", "utf-8")
    test_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "zip_files"
        / "test_zip.zip"
    )
    open_file_mock: Mock = mock_open()
    file_manager: IZipFileManager = ZipFileManager()

    # Act
    with patch(
        "allencell_ml_segmenter.utils.zip_file.zip_file_manager.open",
        open_file_mock,
    ):
        file_manager.write_zip_file(test_path, to_write)

    # Assert
    open_file_mock.assert_called_with(test_path, "wb")
    open_file_mock.return_value.write.assert_called_once_with(to_write)


def test_unzip_zipped_file_and_delete_zip() -> None:
    # Arrange
    file_manager: IZipFileManager = ZipFileManager()
    mock_extract_all: Mock = Mock()
    mock_path: Mock = create_autospec(Path)
    mock_path.parent = "parent"
    with patch(
        "allencell_ml_segmenter.utils.zip_file.zip_file_manager.ZipFile"
    ) as zipfile_mock:
        zipfile_mock.return_value.__enter__.return_value.extractall = (
            mock_extract_all
        )

        # Act
        file_manager.unzip_zipped_file_and_delete_zip(mock_path)

    # Assert
    mock_extract_all.assert_called_once_with(mock_path.parent)
    mock_path.unlink.assert_called_once()



from pathlib import Path
from typing import List, Optional

from napari.utils.events import Event as NapariEvent  # type: ignore
from napari.utils.notifications import show_warning  # type: ignore
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QRadioButton,
    QComboBox,
    QGridLayout,
    QLabel,
    QFrame,
    QSizePolicy,
    QListWidgetItem,
)

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.widgets.input_button_widget import (
    InputButton,
    FileInputMode,
)
from allencell_ml_segmenter.widgets.label_with_hint_widget import LabelWithHint

from allencell_ml_segmenter.core.file_input_model import (
    InputMode,
    FileInputModel,
    WidgetMode,
)

from allencell_ml_segmenter.widgets.check_box_list_widget import (
    CheckBoxListWidget,
)

from allencell_ml_segmenter.main.viewer import IViewer
from allencell_ml_segmenter.prediction.service import ModelFileService
from allencell_ml_segmenter.curation.stacked_spinner import StackedSpinner


class FileInputWidget(QWidget):
    """
    A widget containing file inputs for the input to a model prediction.
    """

    TOP_TEXT: str = "On-screen image(s)"
    BOTTOM_TEXT: str = "Image directory"

    def __init__(
        self,
        model: FileInputModel,
        viewer: IViewer,
        service: Optional[ModelFileService],
        include_channel_selection: bool = True,
        mode: WidgetMode = WidgetMode.PREDICTION,
    ):
        super().__init__()
        self._mode: WidgetMode = mode
        self._include_channel_selection: bool = include_channel_selection
        self._model: FileInputModel = model
        self._viewer: IViewer = viewer
        self._service: Optional[ModelFileService] = service
        layout: QVBoxLayout = QVBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum
        )

        # set up napari event listener for layer changes
        # this keeps the layer list in our UI updated as the layers are added/deleted from napari
        self._viewer.subscribe_layers_change_event(
            function=self._update_layer_list
        )

        frame: QFrame = QFrame()
        frame_layout: QVBoxLayout = QVBoxLayout()
        frame.setLayout(frame_layout)

        frame.setObjectName("frame")

        title: LabelWithHint = LabelWithHint("Input image(s)")
        title.set_hint("Image(s) to apply the trained model on")
        if self._mode == WidgetMode.THRESHOLDING:
            title.set_hint("Image(s) for thresholding.")
        title.setObjectName("title")

        layout.addWidget(title)
        layout.addWidget(frame)
        frame_layout.addWidget(QLabel("Select input image(s):"))

        # radiobox for images from napari
        horiz_layout: QHBoxLayout = QHBoxLayout()
        horiz_layout.setSpacing(0)

        self._radio_on_screen: QRadioButton = QRadioButton()
        self._radio_on_screen.toggled.connect(self._on_screen_slot)
        self._radio_on_screen.setObjectName("onScreen")

        horiz_layout.addWidget(self._radio_on_screen)

        question_label: LabelWithHint = LabelWithHint(FileInputWidget.TOP_TEXT)
        question_label.set_hint("Image(s) already opened in napari")
        horiz_layout.addWidget(question_label)

        frame_layout.addLayout(horiz_layout)

        # list of available images on napari
        self._image_list: CheckBoxListWidget = CheckBoxListWidget()
        self._image_list.setEnabled(False)
        self._image_list.setObjectName("imageList")
        self._image_list.checkedSignal.connect(self._process_checked_signal)
        frame_layout.addWidget(self._image_list)

        # radiobox for images from directory
        horiz_layout = QHBoxLayout()
        horiz_layout.setSpacing(0)

        self._radio_directory: QRadioButton = QRadioButton()
        self._radio_directory.toggled.connect(self._from_directory_slot)
        self._radio_directory.setObjectName("radioDirectory")

        horiz_layout.addWidget(self._radio_directory)

        image_dir_layout: QVBoxLayout = QVBoxLayout()

        question_label = LabelWithHint(FileInputWidget.BOTTOM_TEXT)
        question_label.set_hint(
            "All images in the selected directory will be used as input."
        )

        question_label.add_right_space(10)
        image_dir_layout.addWidget(question_label)

        horiz_layout.addLayout(image_dir_layout)

        horiz_layout.addStretch(5)

        self._browse_dir_edit: InputButton = InputButton(
            self._model,
            lambda dir: self._model.set_input_image_path(
                Path(dir), extract_channels=include_channel_selection
            ),
            "Select directory...",
            FileInputMode.DIRECTORY,
        )
        self.input_image_spinner: StackedSpinner = StackedSpinner(
            input_button=self._browse_dir_edit
        )
        self._browse_dir_edit.setEnabled(False)
        horiz_layout.addWidget(self.input_image_spinner)
        frame_layout.addLayout(horiz_layout)

        grid_layout: QGridLayout = QGridLayout()
        if include_channel_selection:
            image_input_label: LabelWithHint = LabelWithHint(
                "Input image's channel"
            )
            image_input_label.set_hint(
                "Select which channel of the input image(s) to apply the trained model on"
            )

            self._channel_select_dropdown: QComboBox = QComboBox()

            self._channel_select_dropdown.setCurrentIndex(-1)
            self._channel_select_dropdown.currentIndexChanged.connect(
                self._model.set_image_input_channel_index
            )
            self._channel_select_dropdown.setEnabled(False)
            # Event to trigger combobox populate when we know the number of channels
            self._model.subscribe(
                Event.ACTION_FILEINPUT_MAX_CHANNELS_SET,
                self,
                self._populate_input_channel_combobox,
            )
            # Event to set combobox text to 'loading' when we begin extracting channels
            self._model.subscribe(
                Event.ACTION_FILEINPUT_EXTRACT_CHANNELS,
                self,
                self._set_input_channel_combobox_to_loading,
            )
            grid_layout.addWidget(image_input_label, 0, 0)
            grid_layout.addWidget(self._channel_select_dropdown, 0, 1)

        output_dir_label: LabelWithHint = LabelWithHint("Output directory")
        output_dir_label.set_hint("Directory to save result(s) to")

        self._browse_output_edit: InputButton = InputButton(
            self._model,
            lambda dir: self._model.set_output_directory(dir),
            "Select directory...",
            FileInputMode.DIRECTORY,
        )

        grid_layout.addWidget(output_dir_label, 1, 0)
        grid_layout.addWidget(self._browse_output_edit, 1, 1)

        grid_layout.setColumnStretch(0, 1)
        grid_layout.setColumnStretch(1, 0)

        frame_layout.addLayout(grid_layout)

    def _on_screen_slot(self) -> None:
        """Prohibits usage of non-related input fields if top button is checked."""
        self._reset_channel_combobox()
        self._image_list.setEnabled(True)
        self._browse_dir_edit.setEnabled(False)
        self._update_layer_list()
        self._model.set_input_mode(InputMode.FROM_NAPARI_LAYERS)

    def _from_directory_slot(self) -> None:
        """Prohibits usage of non-related input fields if bottom button is checked."""
        self._reset_channel_combobox()
        self._image_list.setEnabled(False)
        self._browse_dir_edit.setEnabled(True)
        self._model.set_input_mode(InputMode.FROM_PATH)

    def _update_layer_list(self, event: Optional[NapariEvent] = None) -> None:
        existing_selection: List[str] = []
        for i in self._image_list.get_checked_rows():
            q_list_widget_item: Optional[QListWidgetItem] = (
                self._image_list.item(i)
            )
            if q_list_widget_item is not None:
                existing_selection.append(q_list_widget_item.text())

        self._image_list.clear()
        self._reset_channel_combobox()
        for layer in self._viewer.get_layers():
            path_of_layer_image: Optional[Path] = self._viewer.get_source_path(
                layer
            )
            if path_of_layer_image:
                self._image_list.add_item(
                    layer.name, set_checked=layer.name in existing_selection
                )

    def _process_checked_signal(self, row: int, state: Qt.CheckState) -> None:
        if self._model.get_input_mode() == InputMode.FROM_NAPARI_LAYERS:
            selected_indices: List[int] = self._image_list.get_checked_rows()
            selected_paths: List[Path] = []
            for i in selected_indices:
                source_path: Optional[Path] = self._viewer.get_source_path(
                    self._viewer.get_layers_nonthreshold()[i]
                )
                if source_path is not None:
                    selected_paths.append(source_path)

            # this will preserve the invariant: the options in the dropdown will be equal
            # to the number of channels in at least one of the selected images (or empty if no images selected)
            if state == Qt.CheckState.Checked:
                if len(selected_paths) <= 10:
                    # only extract if it's the only one checked and it's just been checked;
                    # otherwise we assume they are checking images with same number of channels
                    self._model.set_selected_paths(
                        selected_paths,
                        extract_channels=len(selected_paths) == 1
                        and self._include_channel_selection,
                    )
                else:
                    show_warning(
                        "Cannot predict on > 10 images from the viewer"
                    )
                    qlist_item: Optional[QListWidgetItem] = (
                        self._image_list.item(row)
                    )
                    if qlist_item is not None:
                        qlist_item.setCheckState(Qt.CheckState.Unchecked)
            elif state == Qt.CheckState.Unchecked:
                # could have unselected the img we got channels from originally, so need to re-extract
                # as long as there are still some images selected
                if (
                    len(selected_indices) == 0
                    and self._include_channel_selection
                ):
                    if self._service:
                        self._service.stop_channel_extraction()  # stop so combobox doesn't reset after thread is finished
                    self._reset_channel_combobox()

                self._model.set_selected_paths(
                    selected_paths,
                    extract_channels=len(selected_paths) > 0
                    and self._include_channel_selection,
                )

    def _reset_channel_combobox(self) -> None:
        if self._include_channel_selection:
            self._channel_select_dropdown.clear()
            self._channel_select_dropdown.setPlaceholderText("")
            self._channel_select_dropdown.setCurrentIndex(-1)
            self._channel_select_dropdown.setEnabled(False)

    def _set_input_channel_combobox_to_loading(
        self, event: Optional[Event] = None
    ) -> None:
        if self._include_channel_selection:
            self.input_image_spinner.start()
            self._channel_select_dropdown.clear()
            self._channel_select_dropdown.setPlaceholderText(
                "loading channels..."
            )
            self._channel_select_dropdown.setCurrentIndex(-1)
            self._channel_select_dropdown.setEnabled(False)

    def _populate_input_channel_combobox(
        self, event: Optional[Event] = None
    ) -> None:
        if self._include_channel_selection:
            self.input_image_spinner.stop()
            channels_in_image: Optional[int] = self._model.get_max_channels()
            max_channels: Optional[int] = self._model.get_max_channels()
            self._reset_channel_combobox()
            if (
                channels_in_image is not None
                and channels_in_image > 0
                and max_channels is not None
            ):
                values_range: List[str] = [str(i) for i in range(max_channels)]
                self._channel_select_dropdown.setPlaceholderText(
                    "select a channel index"
                )

                self._channel_select_dropdown.addItems(values_range)
                self._channel_select_dropdown.setEnabled(True)
            else:
                self._channel_select_dropdown.setPlaceholderText(
                    "no channels to select"
                )



from pathlib import Path

import allencell_ml_segmenter
from allencell_ml_segmenter._tests.fakes.fake_subscriber import FakeSubscriber
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.file_input_model import (
    FileInputModel,
    InputMode,
)


def test_set_selected_paths_no_extract_channels() -> None:
    # Arrange
    file_input_model: FileInputModel = FileInputModel()
    dummy_subscriber: FakeSubscriber = FakeSubscriber()
    file_input_model.subscribe(
        Event.ACTION_FILEINPUT_EXTRACT_CHANNELS,
        dummy_subscriber,
        dummy_subscriber.handle,
    )

    # Act
    file_input_model.set_selected_paths([Path()], False)

    # Assert nothing happened

    assert not dummy_subscriber.was_handled(
        Event.ACTION_FILEINPUT_EXTRACT_CHANNELS
    )


def test_set_selected_paths_no_paths() -> None:
    # Arrange
    file_input_model: FileInputModel = FileInputModel()
    dummy_subscriber: FakeSubscriber = FakeSubscriber()
    file_input_model.subscribe(
        Event.ACTION_FILEINPUT_EXTRACT_CHANNELS,
        dummy_subscriber,
        dummy_subscriber.handle,
    )

    # Act
    file_input_model.set_selected_paths(None, True)

    # Assert nothing happened

    assert not dummy_subscriber.was_handled(
        Event.ACTION_FILEINPUT_EXTRACT_CHANNELS
    )


def test_set_selected_paths_dispatched() -> None:
    # Arrange
    file_input_model: FileInputModel = FileInputModel()
    dummy_subscriber: FakeSubscriber = FakeSubscriber()
    file_input_model.subscribe(
        Event.ACTION_FILEINPUT_EXTRACT_CHANNELS,
        dummy_subscriber,
        dummy_subscriber.handle,
    )

    # Act
    file_input_model.set_selected_paths([Path()], True)

    # Assert dispatched
    assert dummy_subscriber.was_handled(
        Event.ACTION_FILEINPUT_EXTRACT_CHANNELS
    )


def test_set_max_channels_no_channel() -> None:
    # Arrange
    file_input_model: FileInputModel = FileInputModel()
    dummy_subscriber: FakeSubscriber = FakeSubscriber()
    file_input_model.subscribe(
        Event.ACTION_FILEINPUT_MAX_CHANNELS_SET,
        dummy_subscriber,
        dummy_subscriber.handle,
    )

    # Act
    file_input_model.set_max_channels(None)

    # Assert nothing happened
    assert not dummy_subscriber.was_handled(
        Event.ACTION_FILEINPUT_MAX_CHANNELS_SET
    )


def test_set_max_channels_dispatch() -> None:
    # Arrange
    file_input_model: FileInputModel = FileInputModel()
    dummy_subscriber: FakeSubscriber = FakeSubscriber()
    file_input_model.subscribe(
        Event.ACTION_FILEINPUT_MAX_CHANNELS_SET,
        dummy_subscriber,
        dummy_subscriber.handle,
    )

    # Act
    file_input_model.set_max_channels(2)

    # Assert nothing happened
    dummy_subscriber.was_handled(Event.ACTION_FILEINPUT_MAX_CHANNELS_SET)


def test_get_input_files_as_list_from_path() -> None:
    """
    Test to see if all paths from a directory are returned as a list
    """
    # ARRANGE
    file_input_model: FileInputModel = FileInputModel()
    file_input_model.set_input_mode(InputMode.FROM_PATH)
    file_input_model.set_input_image_path(
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "img_folder"
    )

    # Act
    files: list[Path] = file_input_model.get_input_files_as_list()

    # Assert
    assert len(files) == 5


def test_get_input_files_as_list_from_viewer() -> None:
    """
    Test to see if all paths from viewer displayed images are returned as a list
    """
    # ARRANGE
    file_input_model: FileInputModel = FileInputModel()
    file_input_model.set_input_mode(InputMode.FROM_NAPARI_LAYERS)
    fake_selected_paths: list[Path] = [Path("fake_path1"), Path("fake_path2")]
    file_input_model.set_selected_paths(fake_selected_paths)

    # Act
    files: list[Path] = file_input_model.get_input_files_as_list()

    # Assert
    assert len(files) == 2
    assert files == fake_selected_paths


def test_get_input_files_as_list_from_no_directory_selected() -> None:
    """
    Test to see if an empty list is returned when no directory is selected
    """
    # ARRANGE
    file_input_model: FileInputModel = FileInputModel()

    # Act
    files: list[Path] = file_input_model.get_input_files_as_list()

    # Assert
    assert len(files) == 0


def test_get_input_files_as_list_from_no_selected_paths() -> None:
    """
    Test to see if an empty list is returned when no layers were selected
    """
    # ARRANGE
    file_input_model: FileInputModel = FileInputModel()

    # Act
    files: list[Path] = file_input_model.get_input_files_as_list()

    # Assert
    assert len(files) == 0


def test_get_input_files_as_list_from_no_selected_paths() -> None:
    """
    Test to see if an empty list is returned when no input mode is selected
    """
    # ARRANGE
    file_input_model: FileInputModel = FileInputModel()

    # Act
    files: list[Path] = file_input_model.get_input_files_as_list()

    # Assert
    assert len(files) == 0



from typing import Optional
from qtpy.QtWidgets import (
    QDialog,
    QWidget,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QPushButton,
)


class InfoDialogBox(QDialog):
    """
    Warning message Widget with a yellow warning sign icon on the left.
    """

    def __init__(self, message: str, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)
        layout: QHBoxLayout = QHBoxLayout()
        self.setLayout(layout)
        layout.addStretch()
        layout.addWidget(QLabel(message))
        layout.addStretch()
        btns = QVBoxLayout()
        self.close_btn = QPushButton("Close")
        self.close_btn.clicked.connect(self._on_close_clicked)
        btns.addWidget(self.close_btn)
        layout.addLayout(btns)

    def _on_close_clicked(self) -> None:
        self.close()



import csv
from pathlib import Path
from typing import List, Dict, Union

from allencell_ml_segmenter._tests.fakes.fake_user_settings import (
    FakeUserSettings,
)
from allencell_ml_segmenter.config.i_user_settings import IUserSettings
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.prediction.model import PredictionModel
import pytest
from unittest.mock import patch, MagicMock, mock_open, call

from allencell_ml_segmenter.services.prediction_service import (
    PredictionService,
)
from allencell_ml_segmenter.core.file_input_model import FileInputModel


@pytest.fixture
def fake_user_settings() -> IUserSettings:
    """
    Fixture for ExperimentsModel Testing
    """
    return FakeUserSettings()


@pytest.mark.skip(reason="cyto_dl mock currently breaks this test on CI")
def test_predict_model() -> None:
    # Arrange
    prediction_model: PredictionModel = PredictionModel()
    experiments_model: ExperimentsModel = ExperimentsModel(
        FakeUserSettings(
            cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
            user_experiments_path=Path(__file__).parent.parent
            / "main"
            / "experiments_home",
        )
    )
    experiments_model.apply_experiment_name("0_exp")
    prediction_service: PredictionService = PredictionService(
        prediction_model, experiments_model
    )

    # Act
    with patch("cyto_dl.api.model.CytoDLModel.predict") as patched_api:
        prediction_model.dispatch(Event.PROCESS_PREDICTION)

    # Assert
    patched_api.assert_called_once()


@pytest.mark.skip(reason="cyto_dl mock currently breaks this test on CI")
def test_predict_model_no_experiment_selected() -> None:
    # Arrange
    prediction_model: PredictionModel = PredictionModel()
    experiments_model: ExperimentsModel = ExperimentsModel(
        FakeUserSettings(
            cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
            user_experiments_path=Path(__file__).parent.parent
            / "main"
            / "experiments_home",
        )
    )
    prediction_service: PredictionService = PredictionService(
        prediction_model, experiments_model
    )

    # Act
    with patch("cyto_dl.api.model.CytoDLModel.predict") as patched_api:
        prediction_model.dispatch(Event.PROCESS_PREDICTION)

    # Assert
    patched_api.assert_not_called()


@pytest.mark.skip(reason="cyto_dl mock currently breaks this test on CI")
def test_predict_model_no_checkpoint_selected() -> None:
    # Arrange
    prediction_model: PredictionModel = PredictionModel()
    experiments_model: ExperimentsModel = ExperimentsModel(
        FakeUserSettings(
            cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
            user_experiments_path=Path(__file__).parent.parent
            / "main"
            / "experiments_home",
        )
    )
    experiments_model.apply_experiment_name("0_exp")
    prediction_service: PredictionService = PredictionService(
        prediction_model, experiments_model
    )

    # Act
    with patch("cyto_dl.api.model.CytoDLModel.predict") as patched_api:
        prediction_model.dispatch(Event.PROCESS_PREDICTION)

    # Assert
    patched_api.assert_not_called()


def test_build_overrides() -> None:
    # Arrange
    file_input_model: FileInputModel = FileInputModel()
    prediction_model: PredictionModel = PredictionModel()
    experiments_model: ExperimentsModel = ExperimentsModel(
        FakeUserSettings(
            cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
            user_experiments_path=Path(__file__).parent.parent
            / "main"
            / "experiments_home",
        )
    )
    experiments_model.apply_experiment_name("one_ckpt_exp")
    prediction_service: PredictionService = PredictionService(
        prediction_model, file_input_model, experiments_model
    )
    file_input_model.set_output_directory(
        Path(__file__).parent.parent
        / "main"
        / "0_exp"
        / "prediction_output_test"
    )
    fake_path: Path = Path("fake_img_path")
    file_input_model.set_input_image_path(fake_path)
    file_input_model.set_image_input_channel_index(3)

    # act
    overrides: Dict[str, Union[str, int, float, bool]] = (
        prediction_service.build_overrides(experiments_model.get_best_ckpt())
    )

    # assert
    # Requried overrides- need these for prediction runs
    assert overrides["test"] == False
    assert overrides["train"] == False
    assert overrides["mode"] == "predict"
    assert overrides["task_name"] == "predict_task_from_app"
    assert overrides["data.path"] == str(fake_path)
    assert overrides["checkpoint.ckpt_path"] == str(
        Path(__file__).parent.parent
        / "main"
        / "experiments_home"
        / "one_ckpt_exp"
        / "checkpoints"
        / "epoch_000.ckpt"
    )

    # optional overrides
    assert overrides["paths.output_dir"] == str(
        Path(__file__).parent.parent
        / "main"
        / "0_exp"
        / "prediction_output_test"
    )
    assert overrides["input_channel"] == 3
    assert overrides["data.columns"] == ["raw", "split"]
    assert overrides["data.split_column"] == "split"


def test_write_csv_for_inputs() -> None:
    # Arrange
    experiments_model: ExperimentsModel = ExperimentsModel(
        FakeUserSettings(
            cyto_dl_home_path=Path(__file__).parent / "cyto_dl_home",
            user_experiments_path=Path(__file__).parent.parent
            / "main"
            / "experiments_home",
        )
    )
    experiments_model.apply_experiment_name("0_exp")
    prediction_model: PredictionModel = PredictionModel()
    prediction_service: PredictionService = PredictionService(
        prediction_model, FileInputModel(), experiments_model
    )
    mock_csv_write = MagicMock(spec=csv.writer)

    # Act
    with patch("builtins.open", mock_open()) as mock_file_open:
        with patch("csv.writer", mock_csv_write):
            prediction_service.write_csv_for_inputs(["image1", "image2"])

    # Assert that CSV.write is called with correct rows
    assert call().writerow(["", "raw", "split"]) in mock_csv_write.mock_calls
    assert (
        call().writerow(["0", "image1", "test"]) in mock_csv_write.mock_calls
    )
    assert (
        call().writerow(["1", "image2", "test"]) in mock_csv_write.mock_calls
    )



from typing import Optional
from napari.layers import Layer  # type: ignore

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.file_input_model import FileInputModel

# Some thresholding constants #
AVAILABLE_AUTOTHRESHOLD_METHODS: list[str] = ["threshold_otsu"]
THRESHOLD_DEFAULT = 120
THRESHOLD_RANGE = (0, 255)


class ThresholdingModel(FileInputModel):
    """
    Stores state relevant to thresholding processes.
    """

    def __init__(self) -> None:
        super().__init__()

        # cyto-dl segmentations should have values between 0 and 255
        self._is_threshold_enabled: bool = False
        self._thresholding_value_selected: int = THRESHOLD_DEFAULT
        self._is_autothresholding_enabled: bool = False
        self._autothresholding_method: str = AVAILABLE_AUTOTHRESHOLD_METHODS[0]
        self._thresholding_layers: list[Layer] = (
            []
        )  # Layers show binary map as data, but contain a metadata
        # key prob_map which contains the probability map that we need to threshold to generate a binary map with
        # the set threshold value.
        self._selected_idx: list[int] = []

    def set_thresholding_value(self, value: int) -> None:
        """
        Set the thresholding value.
        """
        self._thresholding_value_selected = value
        self.dispatch(Event.ACTION_EXECUTE_THRESHOLDING)

    def get_thresholding_value(self) -> int:
        """
        Get the thresholding value.
        """
        return self._thresholding_value_selected

    def set_autothresholding_enabled(self, enable: bool) -> None:
        """
        Set autothresholding enabled.
        """
        self._is_autothresholding_enabled = enable
        if enable:
            self.dispatch(Event.ACTION_EXECUTE_THRESHOLDING)

    def is_autothresholding_enabled(self) -> bool:
        """
        Get autothresholding enabled.
        """
        return self._is_autothresholding_enabled

    def set_autothresholding_method(self, method: str) -> None:
        """
        Set autothresholding method.
        """
        self._autothresholding_method = method
        self.dispatch(Event.ACTION_EXECUTE_THRESHOLDING)

    def get_autothresholding_method(self) -> str:
        """
        Get autothresholding method.
        """
        return self._autothresholding_method

    def set_threshold_enabled(self, enabled: bool) -> None:
        """
        Set threshold specific value.
        """
        self._is_threshold_enabled = enabled
        if enabled:
            self.dispatch(Event.ACTION_EXECUTE_THRESHOLDING)

    def is_threshold_enabled(self) -> bool:
        """
        Get threshold specific value.
        """
        return self._is_threshold_enabled

    def disable_all(self) -> None:
        self.set_threshold_enabled(False)
        self.set_autothresholding_enabled(False)
        self.dispatch(Event.ACTION_THRESHOLDING_DISABLED)

    def dispatch_save_thresholded_images(self) -> None:
        self.dispatch(Event.ACTION_SAVE_THRESHOLDING_IMAGES)

    def set_thresholding_layers(self, layers: list[Layer]) -> None:
        self._thresholding_layers = layers

    def get_thresholding_layers(self) -> list[Layer]:
        return self._thresholding_layers

    def set_selected_idx(self, selected_idx: list[int]) -> None:
        self._selected_idx = selected_idx
        self.dispatch(Event.ACTION_EXECUTE_THRESHOLDING)

    def get_selected_idx(self) -> list[int]:
        return self._selected_idx



from dataclasses import dataclass
import numpy as np
from pathlib import Path
from typing import Optional


@dataclass
class SegmenterLayer:
    name: str


@dataclass
class ShapesLayer(SegmenterLayer):
    data: np.ndarray


@dataclass
class Source:
    path: Optional[Path] = None


@dataclass
class ImageLayer(SegmenterLayer):
    path: Optional[Path] = None
    data: Optional[np.ndarray] = None
    source: Source = Source()
    metadata: Optional[dict] = None


@dataclass
class LabelsLayer(SegmenterLayer):
    metadata: Optional[dict] = None



from enum import Enum
from pathlib import Path
from typing import Callable, Any, Optional

from qtpy.QtWidgets import (
    QWidget,
    QSizePolicy,
    QHBoxLayout,
    QPushButton,
    QLineEdit,
    QFileDialog,
)
from qtpy.QtCore import Qt

from allencell_ml_segmenter._style import Style
from allencell_ml_segmenter.widgets.directory_or_csv_file_dialog import (
    DirectoryOrCSVFileDialog,
)


class FileInputMode(Enum):
    DIRECTORY = "dir"
    FILE = "file"
    DIRECTORY_OR_CSV = "dir_or_csv"


class InputButton(QWidget):
    """
    Compound widget consisting of QLineEdit and QPushButton side-by-side.
    Useful for selecting files and displaying the chosen file path.
    """

    def __init__(
        self,
        model: Any,
        model_set_file_path_function: Callable,
        placeholder: str = "Select file...",
        mode: FileInputMode = FileInputMode.FILE,
    ):
        super().__init__()
        self._default_placeholder = placeholder
        self._model: Any = model

        self.setSizePolicy(
            QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed
        )

        layout: QHBoxLayout = QHBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self._set_path_function: Callable = model_set_file_path_function
        self._mode: FileInputMode = mode

        # text box that will eventually display the chosen file path
        self._text_display: QLineEdit = QLineEdit()
        self._text_display.setPlaceholderText(self._default_placeholder)
        self._text_display.setObjectName("textDisplay")
        self._text_display.setReadOnly(True)

        # button to open file explorer
        self.button: QPushButton = QPushButton("Browse")
        self.button.setObjectName("button")

        # add widgets to layout
        layout.addWidget(
            self._text_display, alignment=Qt.AlignmentFlag.AlignLeft
        )
        layout.addWidget(self.button, alignment=Qt.AlignmentFlag.AlignLeft)

        # connect to slot
        self.button.clicked.connect(self._select_path)
        # connect to stylesheet
        self.setStyleSheet(Style.get_stylesheet("input_button_widget.qss"))

    def _update_path_text(self, path_text: Optional[str]) -> None:
        """
        Displays chosen file or directory path on label portion of input button.
        """
        if path_text is not None:
            self._text_display.setReadOnly(False)
            self._text_display.setText(path_text)
            self._text_display.setReadOnly(True)

            # convert path_text to the appropriate path
            path: Path = Path(path_text)

            self._set_path_function(path)

    def elongate(self, min_width: int) -> None:
        """
        Increases the width of the input button _text_display.
        """
        self._text_display.setMinimumWidth(min_width)

    def _select_path(self) -> None:
        """
        Called whenever an input button is clicked. Takes into account the type of files that can be selected
        and makes a call to _update_path_text if the user selects a compatible file or directory.
        """
        file_path: Optional[str]
        if self._mode == FileInputMode.FILE:
            file_path = QFileDialog.getOpenFileName(
                self,
                "Select a file",
                options=QFileDialog.Option.DontUseNativeDialog
                | QFileDialog.Option.DontUseCustomDirectoryIcons,  # type: ignore
            )[0]
        elif self._mode == FileInputMode.DIRECTORY_OR_CSV:
            custom_dialog: DirectoryOrCSVFileDialog = (
                DirectoryOrCSVFileDialog()
            )
            if custom_dialog.exec_() == QFileDialog.Accepted:
                file_path = custom_dialog.selectedFiles()[0]
            else:
                file_path = None
        else:  # FileInputMode.DIRECTORY
            file_path = QFileDialog.getExistingDirectory(
                self,
                "Select a directory",
                options=QFileDialog.Option.DontUseNativeDialog
                | QFileDialog.Option.DontUseCustomDirectoryIcons
                | QFileDialog.ShowDirsOnly,
            )
        # TODO: boolean field for csv_file_selected (for some service to read it and extract info)?

        if file_path:
            self._update_path_text(file_path)

    def clear_selection(self) -> None:
        self._text_display.clear()
        self._text_display.setPlaceholderText(self._default_placeholder)



from allencell_ml_segmenter.main.main_model import MainModel, ImageType
from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel
from allencell_ml_segmenter.core.task_executor import (
    ITaskExecutor,
    NapariThreadTaskExecutor,
)
from allencell_ml_segmenter.utils.file_writer import IFileWriter, FileWriter
import json
from pathlib import Path
from typing import Optional
from allencell_ml_segmenter.core.event import Event


class MainService:
    def __init__(
        self,
        main_model: MainModel,
        experiments_model: IExperimentsModel,
        task_executor: ITaskExecutor = NapariThreadTaskExecutor.global_instance(),
        file_writer: IFileWriter = FileWriter.global_instance(),
    ):
        self._main_model: MainModel = main_model
        self._experiments_model: IExperimentsModel = experiments_model
        self._task_executor: ITaskExecutor = task_executor
        self._file_writer: IFileWriter = file_writer

        self._main_model.signals.selected_channels_changed.connect(
            self._write_selected_channels
        )
        # TODO: refactor experiments model to use slots + signals
        self._experiments_model.subscribe(
            Event.ACTION_EXPERIMENT_APPLIED, self, self._read_selected_channels
        )

    def _read_selected_channels(self, e: Event) -> None:
        channel_path: Optional[Path] = (
            self._experiments_model.get_channel_selection_path()
        )
        if channel_path is not None:
            self._task_executor.exec(
                lambda: self._read_channel_json(channel_path),
                on_return=self._main_model.set_selected_channels,
            )

    def _read_channel_json(
        self, channel_path: Path
    ) -> dict[ImageType, Optional[int]]:
        if channel_path.exists():
            with open(channel_path, "r") as fr:
                channels: dict[str, Optional[int]] = json.load(fr)
            typed_channels: dict[ImageType, Optional[int]] = {
                ImageType(k): v for k, v in channels.items()
            }
            return typed_channels
        return {
            ImageType.RAW: None,
            ImageType.SEG1: None,
            ImageType.SEG2: None,
        }

    def _write_selected_channels(self) -> None:
        selected_channels: dict[ImageType, Optional[int]] = (
            self._main_model.get_selected_channels()
        )
        # this is a non-critical task, so failing silently is OK--user will just have to manually specify
        # channels during training
        self._task_executor.exec(
            lambda: self._write_channel_json(selected_channels)
        )

    def _write_channel_json(
        self, selected_channels: dict[ImageType, Optional[int]]
    ) -> None:
        jsonified_channels: dict[str, Optional[int]] = {
            k.value: v for k, v in selected_channels.items()
        }
        channel_sel_path: Optional[Path] = (
            self._experiments_model.get_channel_selection_path()
        )
        if channel_sel_path is not None:
            self._file_writer.write_json(
                jsonified_channels,
                channel_sel_path,
            )



import pytest
from pytestqt.qtbot import QtBot
from qtpy.QtWidgets import QListWidgetItem
from qtpy.QtCore import Qt

from allencell_ml_segmenter.widgets.check_box_list_widget import (
    CheckBoxListWidget,
)


@pytest.fixture
def check_box_list_widget(qtbot: QtBot) -> CheckBoxListWidget:
    """
    Returns a CheckBoxListWidget for testing.
    """
    return CheckBoxListWidget()


def test_add_item(check_box_list_widget: CheckBoxListWidget) -> None:
    """
    Tests that add_item() correctly adds a string or QListWidgetItem to the list.
    """
    # ARRANGE/ACT
    check_box_list_widget.add_item("1")
    check_box_list_widget.add_item((QListWidgetItem("2")))

    # ASSERT
    assert check_box_list_widget.count() == 2
    assert check_box_list_widget.item(0).text() == "1"

    # ASSERT - see if string converted to QListWidgetItem
    assert isinstance(check_box_list_widget.item(1), QListWidgetItem)
    assert check_box_list_widget.item(1).text() == "2"


def test_add_item_invalid_throws_error(
    check_box_list_widget: CheckBoxListWidget,
) -> None:
    """
    Tests that adding an item of type other than string or QListWidgetItem throws a TypeError.
    """
    with pytest.raises(TypeError):
        # ACT
        check_box_list_widget.add_item(3)


def test_set_all_state_uniform(
    check_box_list_widget: CheckBoxListWidget,
) -> None:
    """
    Tests that set_all_state() correctly sets all checkboxes to the given state.
    Checkboxes arranged to have a mix of checked and unchecked states.
    """
    # ARRANGE
    check_box_list_widget.add_item("1")
    check_box_list_widget.add_item("2")
    check_box_list_widget.add_item("3")
    check_box_list_widget.item(0).setCheckState(Qt.Checked)
    check_box_list_widget.item(1).setCheckState(Qt.Unchecked)

    # ACT
    check_box_list_widget.set_all_state(Qt.Checked)

    # ASSERT
    for i in range(check_box_list_widget.count()):
        assert check_box_list_widget.item(i).checkState() == Qt.Checked

    # ACT
    check_box_list_widget.set_all_state(Qt.Unchecked)

    # ASSERT
    for i in range(check_box_list_widget.count()):
        assert check_box_list_widget.item(i).checkState() == Qt.Unchecked


def test_toggle_state_uniform(
    check_box_list_widget: CheckBoxListWidget,
) -> None:
    """
    Tests that set_all_state() correctly sets all checkboxes to the given state.
    Checkboxes arranged to have uniform checkstates.
    """
    # ARRANGE
    check_box_list_widget.add_item("1")
    check_box_list_widget.add_item("2")

    # ACT
    check_box_list_widget.set_all_state(Qt.Checked)

    # ASSERT
    for i in range(check_box_list_widget.count()):
        assert check_box_list_widget.item(i).checkState() == Qt.Checked

    # ACT
    check_box_list_widget.set_all_state(Qt.Unchecked)

    # ASSERT
    for i in range(check_box_list_widget.count()):
        assert check_box_list_widget.item(i).checkState() == Qt.Unchecked


def test_get_checked_rows(check_box_list_widget: CheckBoxListWidget) -> None:
    """
    Tests that get_checked_rows() returns the correct indices of checked rows.
    """
    # ARRANGE
    check_box_list_widget.add_item("1")
    check_box_list_widget.add_item("2")
    check_box_list_widget.add_item("3")
    check_box_list_widget.add_item("4")

    # ACT
    check_box_list_widget.item(1).setCheckState(Qt.Checked)
    check_box_list_widget.item(3).setCheckState(Qt.Checked)

    # ASSERT
    assert check_box_list_widget.get_checked_rows() == [1, 3]


def test_get_unchecked_rows(check_box_list_widget: CheckBoxListWidget) -> None:
    """
    Tests that get_unchecked_rows() returns the correct indices of unchecked rows.
    """
    # ARRANGE
    check_box_list_widget.add_item("1")
    check_box_list_widget.add_item("2")
    check_box_list_widget.add_item("3")
    check_box_list_widget.add_item("4")

    # ACT
    check_box_list_widget.item(1).setCheckState(Qt.Checked)
    check_box_list_widget.item(3).setCheckState(Qt.Checked)

    # ASSERT
    assert check_box_list_widget.get_unchecked_rows() == [0, 2]


def test_remove_checked_rows(
    check_box_list_widget: CheckBoxListWidget,
) -> None:
    """
    Tests that remove_checked_rows() correctly removes checked rows.
    """
    # ARRANGE
    check_box_list_widget.add_item("1")
    check_box_list_widget.add_item("2")
    check_box_list_widget.add_item("3")
    check_box_list_widget.add_item("4")

    check_box_list_widget.item(0).setCheckState(Qt.Checked)
    check_box_list_widget.item(2).setCheckState(Qt.Checked)

    # ACT
    check_box_list_widget.remove_checked_rows()

    # ASSERT
    assert check_box_list_widget.count() == 2
    assert check_box_list_widget.item(0).text() == "2"
    assert check_box_list_widget.item(1).text() == "4"


def test_remove_unchecked_rows(
    check_box_list_widget: CheckBoxListWidget,
) -> None:
    """
    Tests that remove_unchecked_rows() correctly removes unchecked rows.
    """
    # ARRANGE
    check_box_list_widget.add_item("1")
    check_box_list_widget.add_item("2")
    check_box_list_widget.add_item("3")
    check_box_list_widget.add_item("4")

    # index 1 and 3 are checked, so 0 and 2 are not
    # initially boxes are unchecked
    check_box_list_widget.item(1).setCheckState(Qt.Checked)
    check_box_list_widget.item(3).setCheckState(Qt.Checked)

    # ACT
    check_box_list_widget.remove_unchecked_rows()

    # ASSERT
    assert check_box_list_widget.count() == 2
    assert check_box_list_widget.item(0).text() == "2"
    assert check_box_list_widget.item(1).text() == "4"



from typing import Optional
from pathlib import Path
import webbrowser

from qtpy.QtWidgets import (
    QDialog,
    QWidget,
    QComboBox,
    QVBoxLayout,
    QPushButton,
)

from allencell_ml_segmenter.core.info_dialog_box import InfoDialogBox
from allencell_ml_segmenter.main.experiments_model import IExperimentsModel
from allencell_ml_segmenter.utils.s3.s3_model_bucket import (
    S3ModelBucket,
)
from allencell_ml_segmenter.utils.s3.s3_bucket_constants import PROD_BUCKET


class ModelDownloadDialog(QDialog):
    def __init__(
        self, parent: Optional[QWidget], experiments_model: IExperimentsModel
    ):
        super().__init__(parent)
        self._experiments_model = experiments_model
        exp_path: Optional[Path] = (
            self._experiments_model.get_user_experiments_path()
        )
        if exp_path is None:
            raise RuntimeError(
                "Cannot download model when experiment dir is unknown"
            )
        self._available_models = S3ModelBucket(
            PROD_BUCKET, exp_path
        ).get_available_models()
        layout: QVBoxLayout = QVBoxLayout()
        self.setLayout(layout)

        self._model_select_dropdown: QComboBox = QComboBox()
        self._model_select_dropdown.setCurrentIndex(-1)
        self._model_select_dropdown.addItems(self._available_models.keys())

        self._download_button: QPushButton = QPushButton("Download")
        self._doc_button: QPushButton = QPushButton(
            "Citation and Documentation"
        )

        self._download_button.clicked.connect(self._download_button_handler)
        self._doc_button.clicked.connect(self._doc_button_handler)

        layout.addWidget(self._doc_button)
        layout.addWidget(self._model_select_dropdown)
        layout.addWidget(self._download_button)

    def _doc_button_handler(self) -> None:
        webbrowser.open(
            "https://allencell.github.io/allencell-segmenter-ml/1_Get-started/4_pretrained-models.html"
        )

    def _download_button_handler(self) -> None:
        selected_model_name: str = str(
            self._model_select_dropdown.currentText()
        )

        self._available_models[selected_model_name].download_model_and_unzip()
        exp_path: Optional[Path] = (
            self._experiments_model.get_user_experiments_path()
        )
        download_complete_message = InfoDialogBox(
            f"Downloaded {selected_model_name} to {exp_path / selected_model_name if exp_path is not None else 'unknown destination'}"
        )
        download_complete_message.exec()
        self._experiments_model.refresh_experiments()  # prevents repeat downloads of model before exiting download dialog



from enum import Enum
from typing import Optional
from copy import deepcopy
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.publisher import Publisher
from allencell_ml_segmenter.core.view import MainWindow
from qtpy.QtCore import QObject, Signal


class ImageType(Enum):
    RAW = "raw"
    SEG1 = "seg1"
    SEG2 = "seg2"


class MainModelSignals(QObject):
    selected_channels_changed: Signal = Signal()


# TODO: once we deprecate custom Publisher, make MainModel a QObject with its own signals instead of composing
MIN_DATASET_SIZE: int = 4


class MainModel(Publisher):
    """
    Main model for this application. Responsible for non-tab-related view switching.
    """

    def __init__(self) -> None:
        super().__init__()
        # Current page of the UI
        self._current_view: Optional[MainWindow] = None
        self._is_new_model: bool = False
        self.signals: MainModelSignals = MainModelSignals()
        self._predictions_in_viewer: bool = (
            False  # Tracks whether predictions are displayed in the viewer
        )

        self._selected_channels: dict[ImageType, Optional[int]] = {
            ImageType.RAW: None,
            ImageType.SEG1: None,
            ImageType.SEG2: None,
        }

    def get_current_view(self) -> Optional[MainWindow]:
        """
        getter/property for current page
        """
        return self._current_view

    def set_current_view(self, view: Optional[MainWindow]) -> None:
        """
        Set the current page in the UI and dispatch a MainEvent
        """
        self._current_view = view
        self.dispatch(Event.ACTION_CHANGE_VIEW)

    def set_new_model(self, is_new_model: bool) -> None:
        """
        Dispatches a new model event
        """
        self._is_new_model = is_new_model
        self.dispatch(Event.ACTION_NEW_MODEL)

    def is_new_model(self) -> bool:
        """
        getter/property for is_new_model
        """
        return self._is_new_model

    def get_selected_channels(self) -> dict[ImageType, Optional[int]]:
        return deepcopy(self._selected_channels)

    def set_selected_channels(
        self, selected_channels: dict[ImageType, Optional[int]]
    ) -> None:
        new_channels: dict[ImageType, Optional[int]] = deepcopy(
            selected_channels
        )
        if any(
            [
                self._selected_channels[k] != new_channels[k]
                for k in new_channels.keys()
            ]
        ):
            self._selected_channels = new_channels
            self.signals.selected_channels_changed.emit()

    def training_complete(self) -> None:
        self.dispatch(Event.PROCESS_TRAINING_COMPLETE)

    def set_predictions_in_viewer(self, predictions_in_viewer: bool) -> None:
        """
        Set if predicted images (probability mappings) are displayed in the viewer.
        """
        self._predictions_in_viewer = predictions_in_viewer

    def are_predictions_in_viewer(self) -> bool:
        """
        Check if predicted images (probability mappings) are displayed in the viewer.
        """
        return self._predictions_in_viewer

    def dispatch_prediction_complete(self) -> None:
        self.dispatch(Event.PROCESS_PREDICTION_COMPLETE)



from pathlib import Path

from napari.layers import Layer, Shapes, Image, Labels  # type: ignore
from napari.layers.shapes.shapes import Mode  # type: ignore
from napari.utils.events import Event as NapariEvent  # type: ignore

from allencell_ml_segmenter.main.i_viewer import IViewer
from allencell_ml_segmenter.main.segmenter_layer import (
    ShapesLayer,
    ImageLayer,
    LabelsLayer,
)
import napari  # type: ignore
from typing import Callable, Optional, Any
import numpy as np


class Viewer(IViewer):
    def __init__(
        self,
        viewer: napari.Viewer,
    ):
        super().__init__()
        self.viewer: napari.Viewer = viewer

    def add_image(self, image: np.ndarray, **kwargs: Any) -> None:
        """
        image: image as a numpy array to add to viewer
        **kwargs: dictionary of kwargs that napari.Viewer.add_image() supports.
        Our plugin uses the `name` and `metadata` kwargs.
        Supports any kwargs defined in https://napari.org/dev/api/napari.Viewer.html#napari.Viewer.add_image
        """
        self.viewer.add_image(image, **kwargs)

    def get_image(self, name: str) -> Optional[ImageLayer]:
        for img in self.get_all_images():
            if img.name == name:
                return img
        return None

    def get_all_images(self) -> list[ImageLayer]:
        imgs: list[ImageLayer] = []
        for l in self.viewer.layers:
            source_path: Optional[Path] = self.viewer.get_source_path(l)
            if isinstance(l, Image) and source_path is not None:
                imgs.append(ImageLayer(l.name, source_path))
            elif isinstance(l, Image):
                imgs.append(ImageLayer(l.name, None))
        return imgs

    def add_shapes(self, name: str, face_color: str, mode: Mode) -> None:
        shapes: Shapes = self.viewer.add_shapes(
            None, name=name, face_color=face_color
        )
        shapes.mode = mode

    def get_shapes(self, name: str) -> Optional[ShapesLayer]:
        for shapes in self.get_all_shapes():
            if shapes.name == name:
                return shapes
        return None

    def get_all_shapes(self) -> list[ShapesLayer]:
        return [
            ShapesLayer(l.name, np.asarray(l.data, dtype=object))
            for l in self.viewer.layers
            if isinstance(l, Shapes)
        ]

    def add_labels(self, data: np.ndarray, **kwargs: Any) -> None:
        """
        data: labels layer as a numpy array to add to viewer
        **kwargs: dictionary of kwargs that napari.Viewer.add_image() supports.
        Our plugin uses the `name` and `metadata` kwargs.
        Supports any kwargs defined in https://napari.org/dev/api/napari.Viewer.html#napari.Viewer.add_image
        """
        self.viewer.add_labels(data, **kwargs)

    def get_labels(self, name: str) -> Optional[LabelsLayer]:
        for labels in self.get_all_labels():
            if labels.name == name:
                return labels
        return None

    def get_all_labels(self) -> list[LabelsLayer]:
        # all items in self.viewer.layers inherit from napari.layers.Layer
        # possible types outlined in https://napari.org/stable/api/napari.layers.html
        return [
            LabelsLayer(l.name)
            for l in self.viewer.layers
            if isinstance(l, Labels)
        ]

    def clear_layers(self) -> None:
        self.viewer.layers.clear()

    def remove_layer(self, name: str) -> bool:
        layer: Optional[Layer] = self._get_layer_by_name(name)
        if layer is not None:
            self.viewer.layers.remove(layer)
            return True
        return False

    def contains_layer(self, name: str) -> bool:
        return self._get_layer_by_name(name) is not None

    def get_layers(self) -> list[Layer]:
        return [l for l in self.viewer.layers]

    def get_layers_nonthreshold(self) -> list[Layer]:
        """
        Get only layers which are not segmentation layers from the viewer.
        These are the layers that do not start with [threshold].
        """
        return [
            l
            for l in self.viewer.layers
            if not l.name.startswith("[threshold]")
        ]

    def subscribe_layers_change_event(
        self, function: Callable[[NapariEvent], None]
    ) -> None:
        self.viewer.layers.events.changed.connect(function)

    def _get_layer_by_name(self, name: str) -> Optional[Layer]:
        layers: list[Layer] = self.get_layers()
        for l in layers:
            if l.name == name:
                return l
        return None

    def get_seg_layers(self) -> list[Layer]:
        """
        Get only segmentation layers (which should be probability mappings) from the viewer.
        These are the layers that start with [seg].
        """
        return [
            layer
            for layer in self.get_layers()
            if layer.name.startswith("[seg]")
        ]

    def insert_binary_map_into_layer(
        self,
        layer: Layer,
        image: np.ndarray,
        remove_seg_layers: bool = False,
    ) -> None:
        """
        Insert a binary mpa image into the viewer.
        If a layer for this binary map image already exists, the new image will replace the old one and refresh the viewer.
        If the layer does not exist, it will be added to the viewer in the correct place (on top of the original raw image:
        index_of_segmentation + 1 in the LayerList)

        :param layer: layer to replace.
        :param image: image to insert
        :param remove_seg_layers: boolean indicating if the layer that is being thresholded is a segmentation layer, and should be removed from the layer once it is updated with the threshold.
        """
        # if threshold has not been previously applied, update name
        if (
            "threshold_applied" not in layer.metadata
            or not layer.metadata["threshold_applied"]
        ):
            layer.name = f"[threshold] {layer.name}"
        layer.data = image
        layer.metadata["threshold_applied"] = True
        layer.refresh()

    def get_source_path(self, layer: Layer) -> Optional[Path]:
        """
        Given a layer, gets that layer's source path- the path to the image on the filesystem.
        """
        # If the image was dragged into napari, we expect a layer.source.path
        if layer.source.path is not None:
            return Path(layer.source.path)
        # if the image was added by the plugin, we expect source_path to be set instead
        if layer.metadata is not None and "source_path" in layer.metadata:
            return Path(layer.metadata["source_path"])
        return None

    def get_all_layers_containing_prob_map(self) -> list[Layer]:
        """
        Get all segmentation labels layers that currently exist in the viewer.
        """
        return [
            layer
            for layer in self.get_layers()
            if "prob_map" in layer.metadata
        ]

    def clear_binary_map_from_layer(self, layer: Layer) -> None:
        """
        We need to keep the layer because it contains the segmentation's probability map.
        So, clear out the binary map by setting it to all zeros
        """
        if "threshold_applied" in layer.metadata:
            layer.metadata["threshold_applied"] = (
                False  # so that we know a threshold has no longer been applied to this image
            )
            layer.name = layer.name.replace(
                "[threshold] ", ""
            )  # remove threshold tag from layer name displayed on viewer
        layer.data = np.zeros(
            layer.data.shape, dtype=bool
        )  # image of type bool
        layer.refresh()



from typing import Dict

from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.core.event import Event


class FakeSubscriber(Subscriber):
    """
    Testing publishers with this fake subscriber class
    using the fake event handler.
    """

    def __init__(self):
        self.handled: Dict[Event, bool] = {}

    def handle(self, event: Event) -> None:
        """
        Fake event handler that indicates that the given event was handled.
        """
        self.handled[event] = True

    def was_handled(self, event: Event) -> bool:
        """
        Returns True if the given event was handled.
        """
        return event in self.handled



from qtpy.QtWidgets import (
    QDialog,
    QWidget,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QPushButton,
)
from typing import Optional


class DialogBox(QDialog):
    """
    Warning message Widget with a yellow warning sign icon on the left.
    """

    def __init__(self, message: str, parent: Optional[QWidget] = None):
        super().__init__(parent=parent)
        self.selection: Optional[bool] = None
        layout: QHBoxLayout = QHBoxLayout()
        self.setLayout(layout)

        self._text = QLabel(message)

        layout.addStretch()
        layout.addWidget(self._text)
        layout.addStretch()
        btns = QVBoxLayout()
        self.yes_btn = QPushButton("Yes")
        self.no_btn = QPushButton("No")
        self.yes_btn.clicked.connect(self.yes_selected)
        self.no_btn.clicked.connect(self.no_selected)
        btns.addWidget(self.yes_btn)
        btns.addWidget(self.no_btn)
        layout.addLayout(btns)

    @property
    def message(self) -> str:
        return self.getMessage()

    def yes_selected(self) -> None:
        self.selection = True
        self.accept()

    def no_selected(self) -> None:
        self.selection = False
        self.reject()

    def setMessage(self, message: str) -> None:
        self._text.setText(message)

    def getMessage(self) -> str:
        return self._text.text()

    def get_selection(self) -> Optional[bool]:
        return self.selection



import pytest
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter.widgets.label_with_hint_widget import LabelWithHint


@pytest.fixture
def label_with_hint(qtbot: QtBot) -> LabelWithHint:
    return LabelWithHint()


def test_set_label_text_updates_label_text(
    label_with_hint: LabelWithHint,
) -> None:
    """
    Tests that the label text is updated when set_label_text is called.
    """
    # ARRANGE
    label_txt: str = "Text goes here"

    # ACT
    label_with_hint.set_label_text(label_txt)

    # ASSERT
    assert label_with_hint._label.text() == label_txt


def test_set_hint_updates_question_mark_tooltip(
    label_with_hint: LabelWithHint,
) -> None:
    """
    Tests that the question mark tooltip is updated when set_hint is called.
    """
    # ARRANGE
    tooltip: str = "Helpful hint"

    # ACT
    label_with_hint.set_hint(tooltip)

    # ASSERT
    assert label_with_hint._question_mark.toolTip() == tooltip


def test_layout_contains_label_and_question_mark(
    label_with_hint: LabelWithHint,
) -> None:
    """
    Tests that the widget layout contains the both label and question mark.
    """
    # ASSERT
    assert label_with_hint.layout().indexOf(label_with_hint._label) != -1
    assert (
        label_with_hint.layout().indexOf(label_with_hint._question_mark) != -1
    )



from dataclasses import dataclass
from pathlib import Path
import numpy as np
from typing import Optional


@dataclass
class ImageData:
    dim_x: Optional[int]
    dim_y: Optional[int]
    dim_z: Optional[int]
    channels: Optional[int]
    np_data: Optional[np.ndarray]
    path: Path



from typing import List, Any

from qtpy.QtWidgets import (
    QLabel,
    QHBoxLayout,
    QVBoxLayout,
    QSizePolicy,
    QComboBox,
    QGridLayout,
    QRadioButton,
    QFrame,
    QDoubleSpinBox,
    QSlider,
)
from qtpy.QtCore import Qt
from magicgui.widgets import FloatSlider
from allencell_ml_segmenter.core.aics_widget import AicsWidget

from allencell_ml_segmenter.prediction.model import PredictionModel
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.widgets.label_with_hint_widget import (
    LabelWithHint,
)


class ModelInputWidget(AicsWidget):
    """
    Handles model input, preprocessing selection, and
    postprocessing selection for prediction.
    """

    TOP_TEXT: str = "none"
    MID_TEXT: str = "simple threshold"
    BOTTOM_TEXT: str = "auto threshold"
    MIN: int = 0
    MAX: int = 100
    STEP: int = 1

    def __init__(self, model: PredictionModel):
        super().__init__()

        self._model: PredictionModel = model

        self.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed
        )

        # instantiate widgets
        self._frame: QFrame = QFrame()
        self._title: LabelWithHint = LabelWithHint("Pre/Post Processing")
        # TODO: hints for widget titles?
        self._preprocessing_label_with_hint: LabelWithHint = LabelWithHint()
        self._method_label: QLabel = QLabel("n/a")
        self._postprocessing_label_with_hint: LabelWithHint = LabelWithHint()

        # radio buttons
        self._top_postproc_button: QRadioButton = QRadioButton()
        self._mid_postproc_button: QRadioButton = QRadioButton()
        self._bottom_postproc_button: QRadioButton = QRadioButton()

        self._postproc_buttons: List[QRadioButton] = [
            self._top_postproc_button,
            self._mid_postproc_button,
            self._bottom_postproc_button,
        ]

        # labels for the radio buttons
        self._top_postproc_label: QLabel = QLabel(ModelInputWidget.TOP_TEXT)
        self._mid_postproc_label: LabelWithHint = LabelWithHint(
            ModelInputWidget.MID_TEXT
        )
        self._bottom_postproc_label: LabelWithHint = LabelWithHint(
            ModelInputWidget.BOTTOM_TEXT
        )

        self._postproc_labels: List[Any] = [
            self._top_postproc_label,
            self._mid_postproc_label,
            self._bottom_postproc_label,
        ]

        # input fields corresponding to radio buttons & their labels
        self._simple_thresh_slider: FloatSlider = FloatSlider(
            min=ModelInputWidget.MIN,
            max=ModelInputWidget.MAX,
            step=ModelInputWidget.STEP,
            readout=True,
        )
        self._lower_bound: QLabel = QLabel(str(ModelInputWidget.MIN))
        self._upper_bound: QLabel = QLabel(str(ModelInputWidget.MAX))

        self._auto_thresh_selection: QComboBox = QComboBox()

        self._postprocessing_selections: List[Any] = [
            self._simple_thresh_slider,
            self._auto_thresh_selection,
        ]

        # finish default set-up
        self._call_setters()
        self._build_layouts()
        self._configure_slots()

    def handle_event(self, event: Event) -> None:
        pass

    def _top_postproc_button_slot(self) -> None:
        """
        Prohibits usage of non-related input fields if "none" button is checked.
        """
        # disable middle
        self._simple_thresh_slider.native.setEnabled(False)

        # disable bottom
        self._auto_thresh_selection.setEnabled(False)

    def _mid_postproc_button_slot(self) -> None:
        """
        Prohibits usage of non-related input fields if top button is checked.
        """
        self._simple_thresh_slider.native.setEnabled(True)
        self._auto_thresh_selection.setEnabled(False)

    def _bottom_postproc_button_slot(self) -> None:
        """
        Prohibits usage of non-related input fields if bottom button is checked.
        """
        self._simple_thresh_slider.native.setEnabled(False)
        self._auto_thresh_selection.setEnabled(True)

    def _call_setters(self) -> None:
        """
        Sets pertinent default values for all widget fields.
        """
        # title frame + title
        self._frame.setObjectName("frame")
        self._title.setObjectName("title")

        # preprocessing label + hint
        self._preprocessing_label_with_hint.set_label_text(
            "Preprocessing method"
        )
        self._preprocessing_label_with_hint.set_hint(
            "Image processing steps (e.g. normalization, smoothing) to run prior to segmentation. NOTE: this should be consistent with the image preprocessing used for training and only changed in rare circumstances."
        )

        # styling for label for preprocessing method
        self._method_label.setObjectName("methodLabel")

        # postprocessing label + hint
        self._postprocessing_label_with_hint.set_label_text(
            "Postprocessing methods"
        )
        self._postprocessing_label_with_hint.set_hint(
            "Method for turning model-predicted probabilities into binary masks."
        )

        # Radio label hints
        self._mid_postproc_label.set_hint(
            "Threshold predicted probabilities at a specific value."
        )
        self._bottom_postproc_label.set_hint(
            "Automatically choose a threshold at which to binarize predicted probabilities."
        )

        # slider + spinbox
        slider: QSlider = (
            self._simple_thresh_slider.native.layout().itemAt(0).widget()
        )
        slider.setObjectName("slider")

        spinbox: QDoubleSpinBox = (
            self._simple_thresh_slider.native.layout().itemAt(1).widget()
        )
        spinbox.setObjectName("spinbox")
        spinbox.setStyleSheet(
            """
            #spinbox {
                padding: 0px;
                margin-bottom: 5px;
                background-color: #414851;
            }
        """
        )

        # slider's bounds
        self._lower_bound.setObjectName("lowerBound")
        self._upper_bound.setObjectName("upperBound")

        # set default values for input fields
        self._auto_thresh_selection.addItems(
            [
                "isodata",
                "li",
                "local",
                "mean",
                "minimum",
                "multiotsu",
                "niblack",
                "otsu",
                "savola",
                "triangle",
                "yen",
                "try all",
            ]
        )

        # set up disappearing placeholder text
        self._auto_thresh_selection.setEditable(True)
        self._auto_thresh_selection.setCurrentIndex(-1)

        # TODO: check that the user has selected an option other than the placeholder text when "run" is pressed
        # TODO: also check that one of the three radio buttons has been selected
        self._auto_thresh_selection.setPlaceholderText("select a method")
        self._auto_thresh_selection.setEditable(False)

        # prohibit input until a radio button is selected
        for selection in self._postprocessing_selections:
            if isinstance(selection, FloatSlider):
                selection.native.setEnabled(False)
            else:
                selection.setEnabled(False)

    def _build_layouts(self) -> None:
        """
        Places previously instantiated widgets into respective layouts.
        """
        # initial set-up
        layout: QVBoxLayout = QVBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)

        frame_layout: QVBoxLayout = QVBoxLayout()
        self._frame.setLayout(frame_layout)

        layout.addWidget(self._title)
        layout.addWidget(self._frame)

        # horizontal layout containing widgets related to preprocessing
        preprocessing_layout: QHBoxLayout = QHBoxLayout()
        preprocessing_layout.setSpacing(0)

        preprocessing_layout.addWidget(
            self._preprocessing_label_with_hint,
            alignment=Qt.AlignmentFlag.AlignLeft,
        )
        preprocessing_layout.addWidget(
            self._method_label, alignment=Qt.AlignmentFlag.AlignRight
        )

        # grid layout containing widgets related to postprocessing
        grid_layout: QGridLayout = QGridLayout()
        grid_layout.setSpacing(0)

        # slider bounds
        horiz_layout: QHBoxLayout = QHBoxLayout()
        horiz_layout.addWidget(self._lower_bound)
        horiz_layout.addWidget(self._upper_bound, Qt.AlignmentFlag.AlignLeft)
        horiz_layout.setSpacing(0)

        # slider + bounds
        vert_layout: QVBoxLayout = QVBoxLayout()
        vert_layout.addLayout(horiz_layout)
        vert_layout.addWidget(self._simple_thresh_slider.native)

        # add all pertinent widgets to the grid
        for idx, button in enumerate(self._postproc_buttons):
            grid_layout.addWidget(button, idx, 0)
        for idx, label in enumerate(self._postproc_labels):
            grid_layout.addWidget(label, idx, 1)

        grid_layout.addLayout(vert_layout, idx + 1, 2)
        grid_layout.addWidget(self._auto_thresh_selection, idx + 1, 2)

        grid_layout.setColumnStretch(1, 1)

        # add inner widgets and layouts to overarching layout
        frame_layout.addLayout(preprocessing_layout)
        frame_layout.addWidget(
            self._postprocessing_label_with_hint,
            alignment=Qt.AlignmentFlag.AlignLeft,
        )
        frame_layout.addLayout(grid_layout)

    def _configure_slots(self) -> None:
        """
        Connects widgets to their respective event handlers.
        """
        # connect radio buttons to slots
        self._top_postproc_button.toggled.connect(
            self._top_postproc_button_slot
        )
        self._mid_postproc_button.toggled.connect(
            self._mid_postproc_button_slot
        )
        self._bottom_postproc_button.toggled.connect(
            self._bottom_postproc_button_slot
        )



from typing import Optional

from allencell_ml_segmenter.core.file_input_model import (
    InputMode,
    WidgetMode,
)
from allencell_ml_segmenter.core.file_input_widget import FileInputWidget
from allencell_ml_segmenter.main.i_viewer import IViewer
from qtpy.QtCore import Qt

from napari.utils.events import Event as NapariEvent  # type: ignore

from allencell_ml_segmenter.thresholding.thresholding_model import (
    ThresholdingModel,
)


class ThresholdingFileInputWidget(FileInputWidget):
    """
    Widget containing a list of prediction results that are selectable for thresholding
    """

    def __init__(
        self,
        model: ThresholdingModel,
        viewer: IViewer,
    ):
        super().__init__(model, viewer, None, False, WidgetMode.THRESHOLDING)
        self._model: ThresholdingModel = model

    def _update_layer_list(self, event: Optional[NapariEvent] = None) -> None:
        previous_selections: list[int] = self._image_list.get_checked_rows()
        self._image_list.clear()
        self._model.set_thresholding_layers(
            self._viewer.get_all_layers_containing_prob_map()
        )
        for idx, prediction_output_layer in enumerate(
            self._model.get_thresholding_layers()
        ):
            self._image_list.add_item(
                prediction_output_layer.name,
                set_checked=idx in previous_selections,
            )

    def _process_checked_signal(self, row: int, state: Qt.CheckState) -> None:
        if self._model.get_input_mode() == InputMode.FROM_NAPARI_LAYERS:
            self._model.set_selected_idx(self._image_list.get_checked_rows())



from abc import ABC, abstractmethod
from pathlib import Path
from qtpy.QtWidgets import QWidget
from typing import Optional


class IUserSettings(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def get_cyto_dl_home_path(self) -> Path:
        pass

    @abstractmethod
    def get_user_experiments_path(self) -> Optional[Path]:
        pass

    @abstractmethod
    def set_user_experiments_path(self, path: Path) -> None:
        pass

    @abstractmethod
    def prompt_for_user_experiments_home(self, parent: QWidget) -> None:
        pass

    @abstractmethod
    def display_change_user_experiments_home(self, parent: QWidget) -> None:
        pass



from pathlib import Path
from watchdog.observers.api import BaseObserver
from watchdog.observers import Observer
from allencell_ml_segmenter.core.progress_tracker import ProgressTracker
from allencell_ml_segmenter.training.metrics_csv_event_handler import (
    MetricsCSVEventHandler,
)
from allencell_ml_segmenter.training.cache_dir_event_handler import (
    CacheDirEventHandler,
)
from typing import Optional


class TrainingProgressTracker(ProgressTracker):
    """
    A MetricsCSVProgressTracker measures progress by observing a metrics CSV file
    produced by cyto-dl and using the number of epochs written to it as a
    measure of progress. Relies heavily on current cyto-dl file logging procedure.
    """

    def __init__(
        self,
        csv_path: Path,
        cache_path: Path,
        num_epochs: int,
        total_files: int,
        version_number: int,
    ):
        """
        :param csv_path: path to cyto-dl csv directory for an experiment
        :param cache_path: path to cache directory for an experiment
        :param num_epochs: number of epochs that will be recorded in the csv
        :param total_files: total number of unique files used for training
        :param version_number: experiment version to track
        """
        # set to 0, 0 to initially have a spinning progress bar
        super().__init__(
            progress_minimum=0,
            progress_maximum=0,
            label_text=f"Files cached: 0 / {total_files}",
        )
        self._num_epochs = num_epochs

        self._csv_path: Path = csv_path
        # these paths will be created by cyto-dl anyways, but we need to make
        # sure they exist before we try to set an observer on them
        if not csv_path.exists():
            csv_path.mkdir(parents=True)

        self._cache_path: Path = cache_path
        if not cache_path.exists():
            cache_path.mkdir(parents=True)

        self._target_path: Path = (
            csv_path / f"version_{version_number}" / "metrics.csv"
        )
        self._observer: Optional[BaseObserver] = None
        self._total_files: int = total_files

    def _set_progress(self, progress: int) -> None:
        # on the first call to _set_progress, we expect the progress bar to
        # be in a spinning state (max = 0), we want to change that to the real max
        # to show deterministic progress here
        if self.get_progress_maximum() != self._num_epochs:
            self.set_progress_maximum(self._num_epochs)
        self.set_progress(progress)

    # override
    def start_tracker(self) -> None:
        self.stop_tracker()
        self._observer = Observer()
        csv_handler: MetricsCSVEventHandler = MetricsCSVEventHandler(
            self._target_path, self._set_progress, self.set_label_text
        )
        self._observer.schedule(
            csv_handler, path=str(self._csv_path.resolve()), recursive=True
        )
        cache_handler: CacheDirEventHandler = CacheDirEventHandler(
            self._set_cache_progress_text
        )
        self._observer.schedule(
            cache_handler, path=str(self._cache_path.resolve()), recursive=True
        )
        self._observer.start()

    # override
    def stop_tracker(self) -> None:
        if self._observer:
            self._observer.stop()

    def _set_cache_progress_text(self, num_cached: int) -> None:
        self.set_label_text(
            f"Files cached: {num_cached} / {self._total_files}"
        )
